import logging
import warnings
from collections import OrderedDict

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

from functools import partial

import torch
import torchvision
from torch import nn

from ... import regularizers
from ..activations import AdaptiveELU
from ..affine import Bias2DLayer, Scale2DLayer
from ..attention import AttentionConv
from ..conv import DepthSeparableConv2d
from ..hermite import (
    HermiteConv2D,
    RotationEquivariantBatchNorm2D,
    RotationEquivariantBias2DLayer,
    RotationEquivariantScale2DLayer,
)
from ..squeeze_excitation import SqueezeExcitationBlock
from .base import Core

logger = logging.getLogger(__name__)


class Stacked2dCore(Core, nn.Module):
    """
    An instantiation of the Core base class. Made up of layers layers of nn.sequential modules.
    Allows for the flexible implementations of many different architectures, such as convolutional layers,
    or self-attention layers.
    """

    def __init__(
        self,
        input_channels,
        hidden_channels,
        input_kern,
        hidden_kern,
        layers=3,
        gamma_hidden=0,
        gamma_input=0.0,
        skip=0,
        stride=1,
        final_nonlinearity=True,
        elu_shift=(0, 0),
        bias=True,
        momentum=0.1,
        pad_input=True,
        hidden_padding=None,
        batch_norm=True,
        batch_norm_scale=True,
        independent_bn_bias=True,
        hidden_dilation=1,
        laplace_padding=0,
        input_regularizer="LaplaceL2",
        stack=None,
        use_avg_reg=True,
        depth_separable=False,
        attention_conv=False,
        linear=False,
    ):
        """
        Args:
            input_channels:     Integer, number of input channels as in
            hidden_channels:    Number of hidden channels (i.e feature maps) in each hidden layer
            input_kern:     kernel size of the first layer (i.e. the input layer)
            hidden_kern:    kernel size of each hidden layer's kernel
            layers:         number of layers
            gamma_hidden:   regularizer factor for group sparsity
            gamma_input:    regularizer factor for the input weights (default: LaplaceL2, see neuralpredictors.regularizers)
            skip:           Adds a skip connection
            stride:         stride of the 2d conv layer.
            final_nonlinearity: Boolean, if true, appends an ELU layer after the last BatchNorm (if BN=True)
            elu_shift: a tuple to shift the elu in the following way: Elu(x - elu_xshift) + elu_yshift
            bias:           Adds a bias layer.
            momentum:       momentum in the batchnorm layer.
            pad_input:      Boolean, if True, applies zero padding to all convolutions
            hidden_padding: int or list of int. Padding for hidden layers. Note that this will apply to all the layers
                            except the first (input) layer.
            batch_norm:     Boolean, if True appends a BN layer after each convolutional layer
            batch_norm_scale: If True, a scaling factor after BN will be learned.
            independent_bn_bias:    If False, will allow for scaling the batch norm, so that batchnorm
                                    and bias can both be true. Defaults to True.
            hidden_dilation:    If set to > 1, will apply dilated convs for all hidden layers
            laplace_padding: Padding size for the laplace convolution. If padding = None, it defaults to half of
                the kernel size (recommended). Setting Padding to 0 is not recommended and leads to artefacts,
                zero is the default however to recreate backwards compatibility.
            input_regularizer:  String that must match one of the regularizers in ..regularizers
            stack:        Int or iterable. Selects which layers of the core should be stacked for the readout.
                            default value will stack all layers on top of each other.
                            Implemented as layers_to_stack = layers[stack:]. thus:
                                stack = -1 will only select the last layer as the readout layer.
                                stack of -2 will read out from the last two layers.
                                And stack of 1 will read out from layer 1 (0 indexed) until the last layer.
            use_avg_reg:    bool. Whether to use the averaged value of regularizer(s) or the summed.
            depth_separable: Boolean, if True, uses depth-separable convolutions in all layers after the first one.
            attention_conv: Boolean, if True, uses self-attention instead of convolution for all layers after the first one.
            linear:         Boolean, if True, removes all nonlinearities

            To enable learning batch_norms bias and scale independently, the arguments bias, batch_norm and batch_norm_scale
            work together: By default, all are true. In this case there won't be a bias learned in the convolutional layer, but
            batch_norm will learn both its bias and scale. If batch_norm is false, but bias true, a bias will be learned in the
            convolutional layer. If batch_norm and bias are true, but batch_norm_scale is false, batch_norm won't have learnable
            parameters and a BiasLayer will be added after the batch_norm layer.
        """

        if depth_separable and attention_conv:
            raise ValueError("depth_separable and attention_conv can not both be true")

        super().__init__()
        regularizer_config = (
            dict(padding=laplace_padding, kernel=input_kern)
            if input_regularizer == "GaussianLaplaceL2"
            else dict(padding=laplace_padding)
        )
        self._input_weights_regularizer = regularizers.__dict__[input_regularizer](**regularizer_config)
        self.num_layers = layers
        self.gamma_input = gamma_input
        self.gamma_hidden = gamma_hidden
        self.input_channels = input_channels
        self.hidden_channels = hidden_channels
        self.skip = skip
        self.stride = stride
        self.use_avg_reg = use_avg_reg
        if use_avg_reg:
            warnings.warn("The averaged value of regularizer will be used.", UserWarning)
        self.hidden_padding = hidden_padding
        self.input_kern = input_kern
        self.hidden_kern = hidden_kern
        self.laplace_padding = laplace_padding
        self.hidden_dilation = hidden_dilation
        self.final_nonlinearity = final_nonlinearity
        self.elu_xshift, self.elu_yshift = elu_shift
        self.bias = bias
        self.momentum = momentum
        self.pad_input = pad_input
        self.batch_norm = batch_norm
        self.batch_norm_scale = batch_norm_scale
        self.independent_bn_bias = independent_bn_bias
        if stack is None:
            self.stack = range(self.num_layers)
        else:
            self.stack = [*range(self.num_layers)[stack:]] if isinstance(stack, int) else stack
        self.linear = linear

        if depth_separable:
            self.conv_layer_name = "ds_conv"
            self.ConvLayer = DepthSeparableConv2d
            self.ignore_group_sparsity = True
        elif attention_conv:
            # TODO: check if name attention_conv is backwards compatible
            self.conv_layer_name = "attention_conv"
            self.ConvLayer = self.AttentionConvWrapper
            self.ignore_group_sparsity = True
        else:
            self.conv_layer_name = "conv"
            self.ConvLayer = nn.Conv2d
            self.ignore_group_sparsity = False

        if (self.ignore_group_sparsity) and (gamma_hidden > 0):
            warnings.warn(
                "group sparsity can not be calculated for the requested conv type. Hidden channels will not be regularized and gamma_hidden is ignored."
            )
        self.set_batchnorm_type()
        self.features = nn.Sequential()
        self.add_first_layer()
        self.add_subsequent_layers()
        self.initialize()

    def set_batchnorm_type(self):
        self.batchnorm_layer_cls = nn.BatchNorm2d
        self.bias_layer_cls = Bias2DLayer
        self.scale_layer_cls = Scale2DLayer

    def add_bn_layer(self, layer):
        if self.batch_norm:
            if self.independent_bn_bias:
                layer["norm"] = self.batchnorm_layer_cls(self.hidden_channels, momentum=self.momentum)
            else:
                layer["norm"] = self.batchnorm_layer_cls(
                    self.hidden_channels, momentum=self.momentum, affine=self.bias and self.batch_norm_scale
                )
                if self.bias:
                    if not self.batch_norm_scale:
                        layer["bias"] = self.bias_layer_cls(self.hidden_channels)
                elif self.batch_norm_scale:
                    layer["scale"] = self.scale_layer_cls(self.hidden_channels)

    def add_activation(self, layer):
        if self.linear:
            return
        if len(self.features) < self.num_layers - 1 or self.final_nonlinearity:
            layer["nonlin"] = AdaptiveELU(self.elu_xshift, self.elu_yshift)

    def add_first_layer(self):
        layer = OrderedDict()
        layer["conv"] = nn.Conv2d(
            self.input_channels,
            self.hidden_channels,
            self.input_kern,
            padding=self.input_kern // 2 if self.pad_input else 0,
            bias=self.bias and not self.batch_norm,
        )
        self.add_bn_layer(layer)
        self.add_activation(layer)
        self.features.add_module("layer0", nn.Sequential(layer))

    def add_subsequent_layers(self):
        if not isinstance(self.hidden_kern, Iterable):
            self.hidden_kern = [self.hidden_kern] * (self.num_layers - 1)

        for l in range(1, self.num_layers):
            layer = OrderedDict()
            if self.hidden_padding is None:
                self.hidden_padding = ((self.hidden_kern[l - 1] - 1) * self.hidden_dilation + 1) // 2
            layer[self.conv_layer_name] = self.ConvLayer(
                in_channels=self.hidden_channels if not self.skip > 1 else min(self.skip, l) * self.hidden_channels,
                out_channels=self.hidden_channels,
                kernel_size=self.hidden_kern[l - 1],
                stride=self.stride,
                padding=self.hidden_padding,
                dilation=self.hidden_dilation,
                bias=self.bias,
            )
            self.add_bn_layer(layer)
            self.add_activation(layer)
            self.features.add_module("layer{}".format(l), nn.Sequential(layer))

    class AttentionConvWrapper(AttentionConv):
        def __init__(self, dilation=None, **kwargs):
            """
            Helper class to make an attention conv layer accept input args of a pytorch.nn.Conv2d layer.
            Args:
                dilation: catches this argument from the input args, and ignores it
                **kwargs:
            """
            super().__init__(**kwargs)

    def forward(self, input_):
        ret = []
        for l, feat in enumerate(self.features):
            do_skip = l >= 1 and self.skip > 1
            input_ = feat(input_ if not do_skip else torch.cat(ret[-min(self.skip, l) :], dim=1))
            ret.append(input_)

        return torch.cat([ret[ind] for ind in self.stack], dim=1)

    def laplace(self):
        """
        Laplace regularization for the filters of the first conv2d layer.
        """
        return self._input_weights_regularizer(self.features[0].conv.weight, avg=self.use_avg_reg)

    def group_sparsity(self):
        """
        Sparsity regularization on the filters of all the conv2d layers except the first one.
        """
        ret = 0
        if self.ignore_group_sparsity:
            return ret

        for feature in self.features[1:]:
            ret = ret + feature.conv.weight.pow(2).sum(3, keepdim=True).sum(2, keepdim=True).sqrt().mean()
        return ret / ((self.num_layers - 1) if self.num_layers > 1 else 1)

    def regularizer(self):
        return self.group_sparsity() * self.gamma_hidden + self.gamma_input * self.laplace()

    @property
    def outchannels(self):
        return len(self.features) * self.hidden_channels


class RotationEquivariant2dCore(Stacked2dCore, nn.Module):
    """
    A core built of 2d rotation-equivariant layers. For more info refer to https://openreview.net/forum?id=H1fU8iAqKX.
    """

    def __init__(
        self,
        *args,
        num_rotations=8,
        stride=1,
        upsampling=2,
        rot_eq_batch_norm=True,
        input_regularizer="LaplaceL2norm",
        **kwargs,
    ):
        """
        Args:
            input_channels:     Integer, number of input channels as in
            hidden_channels:    Number of hidden channels (i.e feature maps) in each hidden layer
            input_kern:     kernel size of the first layer (i.e. the input layer)
            hidden_kern:    kernel size of each hidden layer's kernel
            layers:         number of layers
            num_rotations:  number of computed rotations for every feature
            stride:         stride in convolutional layers
            upsampling:     upsampling scale of Hermite filters
            gamma_hidden:   regularizer factor for group sparsity
            gamma_input:    regularizer factor for the input weights (default: LaplaceL2, see neuralpredictors.regularizers)
            final_nonlinearity: Boolean, if true, appends an ELU layer after the last BatchNorm (if BN=True)
            elu_xshift, elu_yshift: final_nonlinearity(x) = Elu(x - elu_xshift) + elu_yshift
            bias:           Adds a bias layer.
            momentum:        momentum in the batchnorm layer.
            pad_input:      Boolean, if True, applies zero padding to all convolutions
            hidden_padding: int or list of int. Padding for hidden layers. Note that this will apply to all the layers
                            except the first (input) layer.
            batch_norm:     Boolean, if True appends a BN layer after each convolutional layer
            batch_norm_scale: If True, a scaling factor after BN will be learned.
            independent_bn_bias:    If False, will allow for scaling the batch norm, so that batchnorm
                                    and bias can both be true. Defaults to True.
            laplace_padding: Padding size for the laplace convolution. If padding = None, it defaults to half of
                the kernel size (recommended). Setting Padding to 0 is not recommended and leads to artefacts,
                zero is the default however to recreate backwards compatibility.
            input_regularizer:  String that must match one of the regularizers in ..regularizers
            stack:        Int or iterable. Selects which layers of the core should be stacked for the readout.
                            default value will stack all layers on top of each other.
                            Implemented as layers_to_stack = layers[stack:]. thus:
                                stack = -1 will only select the last layer as the readout layer.
                                stack of -2 will read out from the last two layers.
                                And stack of 1 will read out from layer 1 (0 indexed) until the last layer.

            use_avg_reg:    bool. Whether to use the averaged value of regularizer(s) or the summed.

            To enable learning batch_norms bias and scale independently, the arguments bias, batch_norm and batch_norm_scale
            work together: By default, all are true. In this case there won't be a bias learned in the convolutional layer, but
            batch_norm will learn both its bias and scale. If batch_norm is false, but bias true, a bias will be learned in the
            convolutional layer. If batch_norm and bias are true, but batch_norm_scale is false, batch_norm won't have learnable
            parameters and a BiasLayer will be added after the batch_norm layer.
        """
        self.num_rotations = num_rotations
        self.stride = stride
        self.upsampling = upsampling
        self.rot_eq_batch_norm = rot_eq_batch_norm
        super().__init__(*args, **kwargs, input_regularizer=input_regularizer)

    def set_batchnorm_type(self):
        if not self.rot_eq_batch_norm:
            self.batchnorm_layer_cls = nn.BatchNorm2d
            self.bias_layer_cls = Bias2DLayer
            self.scale_layer_cls = Scale2DLayer
        else:
            self.batchnorm_layer_cls = partial(RotationEquivariantBatchNorm2D, num_rotations=self.num_rotations)
            self.bias_layer_cls = partial(RotationEquivariantBias2DLayer, num_rotations=self.num_rotations)
            self.scale_layer_cls = partial(RotationEquivariantScale2DLayer, num_rotations=self.num_rotations)

    def add_first_layer(self):
        layer = OrderedDict()
        layer["hermite_conv"] = HermiteConv2D(
            input_features=self.input_channels,
            output_features=self.hidden_channels,
            num_rotations=self.num_rotations,
            upsampling=self.upsampling,
            filter_size=self.input_kern,
            stride=self.stride,
            padding=self.input_kern // 2 if self.pad_input else 0,
            first_layer=True,
        )
        self.add_bn_layer(layer)
        self.add_activation(layer)
        self.features.add_module("layer0", nn.Sequential(layer))

    def add_subsequent_layers(self):
        if not isinstance(self.hidden_kern, Iterable):
            self.hidden_kern = [self.hidden_kern] * (self.num_layers - 1)

        for l in range(1, self.num_layers):
            layer = OrderedDict()

            if self.hidden_padding is None:
                self.hidden_padding = self.hidden_kern[l - 1] // 2

            layer["hermite_conv"] = HermiteConv2D(
                input_features=self.hidden_channels * self.num_rotations,
                output_features=self.hidden_channels,
                num_rotations=self.num_rotations,
                upsampling=self.upsampling,
                filter_size=self.hidden_kern[l - 1],
                stride=self.stride,
                padding=self.hidden_padding,
                first_layer=False,
            )
            self.add_bn_layer(layer)
            self.add_activation(layer)
            self.features.add_module("layer{}".format(l), nn.Sequential(layer))

    def initialize(self):
        self.apply(self.init_conv_hermite)

    @staticmethod
    def init_conv_hermite(m):
        if isinstance(m, HermiteConv2D):
            nn.init.normal_(m.coeffs.data, std=0.1)

    def forward(self, input_):
        ret = []
        for l, feat in enumerate(self.features):
            input_ = feat(input_)
            ret.append(input_)

        return torch.cat([ret[ind] for ind in self.stack], dim=1)

    def laplace(self):
        return self._input_weights_regularizer(self.features[0].conv.weights_all_rotations, avg=self.use_avg_reg)

    def group_sparsity(self):
        ret = 0
        for l in range(1, self.num_layers):
            ret = (
                ret
                + self.features[l]
                .conv.weights_all_rotations.pow(2)
                .sum(3, keepdim=True)
                .sum(2, keepdim=True)
                .sqrt()
                .mean()
            )
        return ret / ((self.num_layers - 1) if self.num_layers > 1 else 1)

    def regularizer(self):
        return self.group_sparsity() * self.gamma_hidden + self.gamma_input * self.laplace()

    @property
    def outchannels(self):
        return len(self.features) * self.hidden_channels * self.num_rotations


class TransferLearningCore(Core, nn.Module):
    """
    Core based on popular image recognition networks from torchvision such as VGG or AlexNet.
    Can be instantiated as random or pretrained. Core is frozen by default, which can be changed with the fine_tune
    argument.
    """

    def __init__(
        self,
        input_channels,
        tl_model_name,
        layers,
        pretrained=True,
        final_batchnorm=True,
        final_nonlinearity=True,
        momentum=0.1,
        fine_tune=False,
        **kwargs,
    ):
        """
        Args:
            input_channels (int): Number of input channels. 1 if greyscale, 3 if RBG
            tl_model_name (str): Name of the image recognition Transfer Learning model. Possible are all models in
            torchvision, i.e. vgg16, alexnet, ...
            layers (int): Number of layers, i.e. after which layer to cut the original network
            pretrained (boolean): Whether to use a randomly initialized or pretrained network
            final_batchnorm (boolean): Whether to add a batch norm after the final conv layer
            final_nonlinearity (boolean): Whether to add a final nonlinearity (ReLU)
            momentum (float): Momentum term for batch norm. Irrelevant if batch_norm=False
            fine_tune (boolean): Whether to clip gradients before this core or to allow training on the core
            **kwargs:
        """
        if kwargs:
            warnings.warn(
                "Ignoring input {} when creating {}".format(repr(kwargs), self.__class__.__name__),
                UserWarning,
            )
        super().__init__()

        self.input_channels = input_channels
        self.momentum = momentum

        # Download model and cut after specified layer
        TL_model = getattr(torchvision.models, tl_model_name)(pretrained=pretrained)
        TL_model_clipped = nn.Sequential(*list(TL_model.features.children())[:layers])
        if not isinstance(TL_model_clipped[-1], nn.Conv2d):
            warnings.warn(
                "Final layer is of type {}, not nn.Conv2d".format(type(TL_model_clipped[-1])),
                UserWarning,
            )

        # Fix pretrained parameters during training
        if not fine_tune:
            for param in TL_model_clipped.parameters():
                param.requires_grad = False

        # Stack model together
        self.features = nn.Sequential()
        self.features.add_module("TransferLearning", TL_model_clipped)
        if final_batchnorm:
            self.features.add_module("OutBatchNorm", nn.BatchNorm2d(self.outchannels, momentum=self.momentum))
        if final_nonlinearity:
            self.features.add_module("OutNonlin", nn.ReLU(inplace=True))

    def forward(self, input_):
        # If model is designed for RBG input but input is greyscale, repeat the same input 3 times
        if self.input_channels == 1 and self.features.TransferLearning[0].in_channels == 3:
            input_ = input_.repeat(1, 3, 1, 1)
        input_ = self.features(input_)
        return input_

    def regularizer(self):
        return 0

    @property
    def outchannels(self):
        """
        Function which returns the number of channels in the output conv layer. If the output layer is not a conv
        layer, the last conv layer in the network is used.

        Returns: Number of output channels
        """
        found_outchannels = False
        i = 1
        while not found_outchannels:
            if "out_channels" in self.features.TransferLearning[-i].__dict__:
                found_outchannels = True
            else:
                i += 1
        return self.features.TransferLearning[-i].out_channels

    def initialize(self):
        logger.warning(
            "Ignoring initialization since the parameters should be acquired from a pretrained model. If you want random weights, set pretrained = False."
        )


class SE2dCore(Stacked2dCore, nn.Module):
    """
    An extension of the Stacked2dCore class. The convolutional layers can be set to be either depth-separable
    (as used in the popular MobileNets) or based on self-attention (as used in Transformer networks).
    Additionally, a SqueezeAndExcitation layer (also called SE-block) can be added after each layer or the n final
    layers. Finally, it is also possible to make this core fully linear, by disabling all nonlinearities.
    This makes it effectively possible to turn a core+readout CNN into a LNP-model.
    """

    def __init__(
        self,
        input_channels,
        hidden_channels,
        input_kern,
        hidden_kern,
        bias=False,
        laplace_padding=None,
        input_regularizer="LaplaceL2norm",
        n_se_blocks=0,
        se_reduction=32,
        use_avg_reg=False,
        **kwargs,
    ):
        """
        Args:
            See Stacked2dCore for all input arguments.

            This core provides the functionality to add Squeeze and Excitation Layers, which can be done through
            these additional arguments:

            se_reduction:   Int, Reduction of channels for global pooling of the Squeeze and Excitation Block.
            n_se_blocks:    Int, number of squeeze and excitation blocks. Inserted from the last layer
                              Examples: layers=4, n_se_blocks=2:
                                => layer0 -> layer1 -> layer2 -> SEblock -> layer3 -> SEblock
        """
        self.n_se_blocks = n_se_blocks
        self.se_reduction = se_reduction

        super().__init__(
            input_channels=input_channels,
            hidden_channels=hidden_channels,
            input_kern=input_kern,
            hidden_kern=hidden_kern,
            bias=bias,
            laplace_padding=laplace_padding,
            input_regularizer=input_regularizer,
            use_avg_reg=use_avg_reg,
            **kwargs,
        )

    def add_subsequent_layers(self):
        if not isinstance(self.hidden_kern, Iterable):
            self.hidden_kern = [self.hidden_kern] * (self.num_layers - 1)

        for l in range(1, self.num_layers):
            layer = OrderedDict()
            if self.hidden_padding is None:
                self.hidden_padding = ((self.hidden_kern[l - 1] - 1) * self.hidden_dilation + 1) // 2
            layer[self.conv_layer_name] = self.ConvLayer(
                in_channels=self.hidden_channels if not self.skip > 1 else min(self.skip, l) * self.hidden_channels,
                out_channels=self.hidden_channels,
                kernel_size=self.hidden_kern[l - 1],
                stride=self.stride,
                padding=self.hidden_padding,
                dilation=self.hidden_dilation,
                bias=self.bias,
            )
            self.add_bn_layer(layer)
            self.add_activation(layer)
            if (self.num_layers - l) <= self.n_se_blocks:
                layer["seg_ex_block"] = SqueezeExcitationBlock(in_ch=self.hidden_channels, reduction=self.se_reduction)
            self.features.add_module("layer{}".format(l), nn.Sequential(layer))

    def regularizer(self):
        return self.gamma_input * self.laplace()



import logging
from collections import Mapping
from pathlib import Path
from zipfile import ZIP_DEFLATED, ZipFile

import h5py
import numpy as np
from tqdm import tqdm

logger = logging.getLogger(__name__)


def zip_dir(zip_name: str, source_dir):
    """
    Zips all files in `source_dir` into a zip file named `zip_name`.
    """
    src_path = Path(source_dir).absolute().resolve(strict=True)
    with ZipFile(zip_name, "w", ZIP_DEFLATED) as zf:
        for file in tqdm(src_path.rglob("*")):
            zf.write(file, file.relative_to(src_path.parent))


def _savenpy(path, val, overwrite):
    if not np.isscalar(val) and len(val.shape) > 0 and val[0].dtype.char == "S":  # convert bytes to unicode
        val = val.astype(str)
    if not path.exists() or overwrite:
        logger.info("Saving", path)
        np.save(path, val)
    else:
        logger.warn("Not overwriting", path)


def convert_movies_h5_dataset_to_folder(
    filename,
    outpath=None,
    overwrite=False,
    data_keys=("inputs", "responses", "behavior", "eye_position"),
):
    """
    Converts an HDF5 dataset used for mouse movie data into a directory structure
    that can be used by the FileTreeDataset.

    Args:
        filename:       filename of the hdf5 file
        outpath:        location of the FileTreeDataset (default .)
        overwrite:      overwrite existing files
    """
    if not isinstance(data_keys, Mapping):
        data_keys = {k: k for k in data_keys}

    h5file = Path(filename)
    outpath = Path(outpath) or h5file.with_suffix("")  # drop extension to form target output

    with h5py.File(filename, "r") as fid:
        for data_key, mapped_name in data_keys.items():
            subpath = outpath / "data/{}".format(mapped_name)
            subpath.mkdir(exist_ok=True, parents=True)
            for key in tqdm(
                fid[data_key].keys(),
                desc="Saving {}".format(data_key),
            ):
                outfile = subpath / "{}.npy".format(key)
                _savenpy(outfile, fid[data_key][key][()], overwrite)

        # map data for trials
        trial_info_keys = [
            "durations",
            "condition_hashes",
            "movie_names",
            "tiers",
            "trial_idx",
            "types",
        ]

        # save tiers
        subpath = outpath / "meta/trials"
        subpath.mkdir(exist_ok=True, parents=True)
        for trial_info_key in trial_info_keys:
            _savenpy(
                subpath / "{}.npy".format(trial_info_key),
                fid[trial_info_key][()],
                overwrite,
            )

        # save nested fields:
        nested_keys = {"neurons": "neurons"}
        for meta_type, target in nested_keys.items():
            for meta_key in fid[meta_type].keys():
                subpath = outpath / "meta/{}".format(target)
                subpath.mkdir(exist_ok=True, parents=True)

                _savenpy(
                    subpath / "{}.npy".format(meta_key),
                    fid[meta_type][meta_key][()],
                    overwrite,
                )

        # save statistics
        def statistics_func(name, node):
            name = name.replace(".", "_").lower()
            if isinstance(node, h5py.Dataset):
                data_file = outpath / "meta/statistics/{}.npy".format(name)
                data_file.parent.mkdir(exist_ok=True, parents=True)
                _savenpy(
                    data_file,
                    node[()],
                    overwrite,
                )

        logger.info("Saving statistics")
        fid["statistics"].visititems(statistics_func)


def convert_static_h5_dataset_to_folder(filename, outpath=None, overwrite=False, ignore_all_behaviors=False):
    """
    Converts a h5 dataset used for mouse data into a directory structure that can be used by the FileTreeDataset.

    Args:
        filename:       filename of the hdf5 file
        outpath:        location of the FileTreeDataset (default .)
        overwrite:      overwrite existing files

    """
    h5file = Path(filename)
    outpath = outpath or (h5file.parent / h5file.stem)

    with h5py.File(filename) as fid:
        attributes = (
            ["images", "responses", "behavior", "pupil_center"] if not ignore_all_behaviors else ["images", "responses"]
        )
        for data_key in attributes:
            subpath = outpath / "data/{}".format(data_key)
            subpath.mkdir(exist_ok=True, parents=True)
            for i, value in tqdm(
                enumerate(fid[data_key]),
                total=fid[data_key].shape[0],
                desc="Saving {}".format(data_key),
            ):
                outfile = subpath / "{}.npy".format(i)
                if not outfile.exists() or overwrite:
                    np.save(outfile, value)

        # save tiers
        for data_key in ["tiers"]:
            subpath = outpath / "meta/trials"
            subpath.mkdir(exist_ok=True, parents=True)
            _savenpy(subpath / "{}.npy".format(data_key), fid[data_key][...], overwrite)

        # save meta info
        for meta_type, target in zip(["item_info", "neurons"], ["trials", "neurons"]):
            for meta_key in fid[meta_type].keys():
                subpath = outpath / "meta/{}".format(target)
                subpath.mkdir(exist_ok=True, parents=True)

                _savenpy(
                    subpath / "{}.npy".format(meta_key),
                    fid[meta_type][meta_key][...],
                    overwrite,
                )

        # save statistics
        def statistics_func(name, node):
            name = name.replace(".", "_").lower()
            if not isinstance(node, h5py.Dataset):
                subpath = outpath / "meta/statistics" / name
                subpath.mkdir(exist_ok=True, parents=True)
            else:
                _savenpy(
                    outpath / "meta/statistics" / "{}.npy".format(name),
                    node[...],
                    overwrite,
                )

        fid["statistics"].visititems(statistics_func)


def load_dict_from_hdf5(filename):
    """
    Given a `filename` of a HDF5 file, opens the file and
    load the entire content as a (nested) dictionary.

    Args:
        filename - name of HDF5 file

    Returns:
        (nested) dictionary corresponding to the content of the HDF5 file.
    """
    with h5py.File(filename, "r") as h5file:
        return recursively_load_dict_contents_from_group(h5file)


def recursively_load_dict_contents_from_group(h5file, path="/"):
    """
    Given a `h5file` h5py object, loads the object at `path`
    as nested dictionary.

    Args:
        h5file - h5py object
        path - Path within the h5py file to load the content of recursively.

    Returns:
        (nested) dictionary corresponding to the content of the HDF5 file at the path.
    """
    ans = {}
    for key, item in h5file[path].items():
        if isinstance(item, h5py.Dataset):
            dtype = item.dtype
            v = item[()]
            if dtype.char == "S":  # convert bytes to univcode
                v = v.astype(str)
            ans[key] = v
        elif isinstance(item, h5py.Group):
            if item.attrs.get("_iterable", False):
                ans[key] = [item[str(i)][()] for i in range(len(item))]
            else:
                ans[key] = recursively_load_dict_contents_from_group(h5file, path + key + "/")
    return ans



import torch
from torch import nn


class Bias2DLayer(nn.Module):
    """
    Bias per channel for a given 2d input.
    """

    def __init__(self, channels, initial=0, **kwargs):
        """
        Args:
            channels (int): number of channels in the input.
            initial (int, optional): intial value. Defaults to 0.
        """
        super().__init__(**kwargs)

        self.bias = torch.nn.Parameter(torch.empty((1, channels, 1, 1)).fill_(initial))

    def forward(self, x):
        return x + self.bias


class Scale2DLayer(nn.Module):
    """
    Scale per channel for a given 2d input.
    """

    def __init__(self, channels, initial=1, **kwargs):
        """
        Args:
            channels (int): number of channels in the input.
            initial (int, optional): intial value. Defaults to 1.
        """
        super().__init__(**kwargs)

        self.scale = torch.nn.Parameter(torch.empty((1, channels, 1, 1)).fill_(initial))

    def forward(self, x):
        return x * self.scale



import numpy as np
import torch
from scipy import ndimage
from skimage import morphology


def rescale(x, in_min, in_max, out_min, out_max):
    in_mean = (in_min + in_max) / 2
    out_mean = (out_min + out_max) / 2
    in_ext = in_max - in_min
    out_ext = out_max - out_min
    gain = out_ext / in_ext
    x_rescaled = (x - in_mean) * gain + out_mean
    return x_rescaled


def apply_mask(mei, mask, pixel_min, pixel_max):
    mei = rescale(mei, pixel_min, pixel_max, -1, 1)
    mask = mask.repeat(len(mei), 1, 1, 1)
    mask = mask.reshape(mei.shape)
    mei_masked = rescale(mei * mask, -1, 1, pixel_min, pixel_max)
    return mei_masked


def create_mask(
    mei,
    zscore_thresh=1.5,
    closing_iters=2,
    gaussian_sigma=1.5,
    return_mask_centroid=False,
):
    """
    Creates a mask for an image using z-score thresholding, binary closing, and Gaussian filtering.

    Parameters:
    - mei (Tensor): The input image (Multi-Exposure Image) to be processed.
    - zscore_thresh (float, optional): The threshold for z-score to identify significant deviations. Default is 1.5.
    - closing_iters (int, optional): Number of iterations for binary closing to close small holes in the mask. Default is 2.
    - gaussian_sigma (float, optional): Sigma value for Gaussian filtering to smooth the mask. Default is 1.5.
    - return_mask_centroid (bool, optional): If True, returns the centroid of the mask along with the mask. Default is False.

    Returns:
    - mask (Tensor): The generated mask for the image.
    - (mask_x, mask_y) (tuple, optional): The centroid of the mask (returned if return_mask_centroid is True).
    """

    # Compute mask centroid
    mei_np = mei.cpu().squeeze().numpy()
    norm_mei_np = (mei_np - mei_np.mean()) / mei_np.std()
    thresholded = (np.abs(norm_mei_np) > zscore_thresh).squeeze()
    closed = ndimage.binary_closing(thresholded, iterations=closing_iters)
    labeled = morphology.label(closed, connectivity=2)
    most_frequent = np.argmax(np.bincount(labeled.ravel())[1:]) + 1
    oneobject = labeled == most_frequent
    hull = morphology.convex_hull_image(oneobject)
    smoothed = ndimage.gaussian_filter(hull.astype(np.float32), sigma=gaussian_sigma)
    mask = torch.Tensor(smoothed)
    # Compute mask centroid
    if return_mask_centroid:
        py, px = (coords.mean() + 0.5 for coords in np.nonzero(hull))
        # mask_y, mask_x = px_y - mask.shape[0] / 2, px_x - mask.shape[1] / 2,
        return mask, {"y": py, "x": px}
    else:
        return mask



import logging
from zipfile import ZipFile

from ...exceptions import DoesNotExistException
from ...transforms import StaticTransform
from ...utils import convert_static_h5_dataset_to_folder, zip_dir
from ..base import FileTreeDatasetBase

logger = logging.getLogger(__name__)


class FileTreeDataset(FileTreeDatasetBase):
    _transform_types = (StaticTransform,)

    @staticmethod
    def initialize_from(filename, outpath=None, overwrite=False):
        """
        Convenience function. See `convert_static_h5_dataset_to_folder` in `.utils`
        """
        convert_static_h5_dataset_to_folder(filename, outpath=outpath, overwrite=overwrite)

    @property
    def img_shape(self):
        return (1,) + self[0].images.shape

    @property
    def n_neurons(self):
        target_group = "responses" if "responses" in self.data_keys else "targets"
        val = self[0]
        if hasattr(val, target_group):
            val = getattr(val, target_group)
        else:
            val = val[target_group]
        return len(val)

    def change_log(self):
        if (self.basepath / "change.log").exists():
            with open(self.basepath / "change.log", "r") as fid:
                logger.info("".join(fid.readlines()))

    def zip(self, filename=None):
        """
        Zips current dataset.
        Args:
            filename:  Filename for the zip. Directory name + zip by default.
        """

        if filename is None:
            filename = str(self.basepath) + ".zip"
        zip_dir(filename, self.basepath)

    def unzip(self, filename, path):
        logger.info(f"Unzipping {filename} into {path}")
        with ZipFile(filename, "r") as zip_obj:
            zip_obj.extractall(path)

    def add_link(self, attr, new_name):
        """
        Add a new dataset that links to an existing dataset.
        For instance `targets` that links to `responses`
        Args:
            attr:       existing attribute such as `responses`
            new_name:   name of the new attribute reference.
        """
        if not (self.basepath / "data/{}".format(attr)).exists():
            raise DoesNotExistException("Link target does not exist")



from collections import OrderedDict
import numpy as np
import torch
from torch import nn

class LinearPerDim(nn.Module):
    def __init__(self, n_dim, in_dim, out_dim, bias=True):
        super().__init__()
        sqrt_k = np.sqrt(1 / in_dim)
        self.weight = nn.Parameter(torch.rand(n_dim, in_dim, out_dim) * 2 * sqrt_k - sqrt_k)
        self.has_bias = False
        if bias:
            self.bias = nn.Parameter(torch.rand(n_dim, 1, out_dim) * 2 * sqrt_k - sqrt_k)
            self.has_bias = True
        
    def forward(self, x):
        out = torch.einsum("dni,dio->dno", x, self.weight)
        return out + self.bias if self.has_bias else out

class CouplingLayer(nn.Module):
    def __init__(self, n_dim, in_dim, untouched_half, affine=False, hidden_features=256, hidden_layers=3, init_identity=True, eps=1e-8):
        super().__init__()
        self.n_dim = n_dim
        self.in_dim = in_dim
        self.untouched_half = untouched_half
        self.affine = affine
        self.init_identity = init_identity
        self.eps = eps
        self.fhi, self.shi = self.dim_indices()
        
        in_dim = len(self.fhi) if self.untouched_half == 'first' else len(self.shi)
        out_dim = len(self.shi) if self.untouched_half == 'first' else len(self.fhi)
        self.t = TranslationNet(n_dim, in_dim, out_dim, hidden_features=hidden_features, hidden_layers=hidden_layers)
        if init_identity:
            self.t.net[-1].weight.data.zero_()
            self.t.net[-1].bias.data.zero_()
        if affine:
            if init_identity:
                self.s = ScaleNet(n_dim, in_dim, out_dim, hidden_features=hidden_features, hidden_layers=hidden_layers)
                self.s.net[-1].weight.data.zero_()
                self.s.net[-1].bias.data.zero_()
        
    def dim_indices(self):
        indices = np.arange(self.in_dim)
        first_half = indices[:self.in_dim//2]
        second_half = indices[self.in_dim//2:]
        return first_half, second_half

    def forward(self, x):
        x1 = x[..., self.fhi]
        x2 = x[..., self.shi]
        
        if self.untouched_half == 'first':
            z1 = x1
            positive_scale = torch.exp(self.s(x1)) if self.affine else 1. 
#             positive_scale = F.elu(self.s(x1)) + 1 + self.eps if self.affine else 1. 
            z2 = x2 * positive_scale + self.t(x1)
        elif self.untouched_half == 'second':
            positive_scale = torch.exp(self.s(x2)) if self.affine else 1.
#             positive_scale = F.elu(self.s(x2)) + 1 + self.eps if self.affine else 1.
            z1 = x1 * positive_scale + self.t(x2)
            z2 = x2
        
        if (len(z1) == 1) and (len(z2) != 1):
            _z1 = z1.repeat(len(z2), 1, 1)
            return torch.cat((_z1, z2), dim=-1)
        elif len(z2) == 1 and (len(z1) != 1):
            _z2 = z2.repeat(len(z1), 1, 1)
            return torch.cat((z1, _z2), dim=-1)
        else:
            return torch.cat((z1, z2), dim=-1)
    
    def inverse(self, z):
        z1 = z[..., self.fhi]
        z2 = z[..., self.shi]
        
        if self.untouched_half == 'first':
            x1 = z1
            positive_scale = torch.exp(self.s(z1)) if self.affine else 1.
            # positive_scale = F.elu(self.s(z1)) + 1 + self.eps if self.affine else 1.
            x2 = (z2 - self.t(z1)) / positive_scale
        elif self.untouched_half == 'second':
            positive_scale = torch.exp(self.s(z2)) if self.affine else 1.
            # positive_scale = F.elu(self.s(z2)) + 1 + self.eps if self.affine else 1.
            x1 = (z1 - self.t(z2)) / positive_scale
            x2 = z2
            
        return torch.cat((x1, x2), dim=-1)
    

class ScaleNet(nn.Module):
    def __init__(self, n_dim, in_dim, out_dim, hidden_features=256, hidden_layers=3):
        super().__init__()
        
        if hidden_layers == 0:
            hidden_features = out_dim
    
        layers = [
            LinearPerDim(
                n_dim,
                in_dim,
                hidden_features,
            )
        ]

        for i in range(hidden_layers):
            layers.extend(
                [
                    nn.ReLU(),
                    LinearPerDim(
                        n_dim, 
                        hidden_features,
                        hidden_features
                        if i < hidden_layers - 1
                        else out_dim,
                    ),
                ]
            )   
        self.net = nn.Sequential(*layers)

    def forward(self, x):
        return self.net(x)

class TranslationNet(nn.Module):
    def __init__(self, n_dim, in_dim, out_dim, hidden_features=256, hidden_layers=3):
        super().__init__()
        
        if hidden_layers == 0:
            hidden_features = out_dim
        
        layers = [
            LinearPerDim(
                n_dim,
                in_dim,
                hidden_features,
            )
        ]

        for i in range(hidden_layers):
            layers.extend(
                [
                    nn.ReLU(),
                    LinearPerDim(
                        n_dim,
                        hidden_features,
                        hidden_features
                        if i < hidden_layers - 1
                        else out_dim,
                    ),
                ]
            )

        self.net = nn.Sequential(*layers)

    def forward(self, x):
        return self.net(x)

class RealNVP(nn.Module):
    def __init__(self, n_dim, n_layers, in_dim, affine=False, init_identity=True, hidden_features=256, hidden_layers=3):
        super().__init__()
        self.n_dim = n_dim
        self.n_layers = n_layers
        self.in_dim = in_dim
        self.layers = nn.ModuleList([
            CouplingLayer(
                n_dim=n_dim,
                in_dim=in_dim, 
                untouched_half=['first', 'second'][i%in_dim], 
                affine=affine,
                init_identity=init_identity,
                hidden_features=hidden_features, 
                hidden_layers=hidden_layers) for i in range(n_layers)
        ])
    
    def inverse(self, z):
        x = z
        for l in self.layers[::-1]:
            x = l.inverse(x)
        
        return x
    
    def forward(self, x):
        z = x
        for l in self.layers:
            z = l.forward(z)

        return z
    
    
class AffineTransformOld(nn.Module):
    def __init__(self, n_neurons, in_dim=2, stochastic=False, init_noise_scale=.1):
        super().__init__()
        self.n_neurons = n_neurons
        self.in_dim = in_dim
        self.stochastic = stochastic
        self.As = nn.Parameter(torch.stack([torch.eye(in_dim)]*n_neurons))
        self.translation = nn.Parameter(torch.zeros(n_neurons, 1, in_dim))

        if stochastic:
            self.init_noise_scale = init_noise_scale
            self._As_noise_scale = nn.Parameter(torch.log(torch.ones(n_neurons, in_dim, in_dim) * init_noise_scale))
            self._translation_noise_scale = nn.Parameter(torch.log(torch.ones(n_neurons, 1, in_dim) * init_noise_scale))
    
    @property
    def As_noise_scale(self):
        return torch.exp(self._As_noise_scale)

    @property
    def translation_noise_scale(self):
        return torch.exp(self._translation_noise_scale)
        
    def forward(self, x, stochastic=None):
        stochastic = stochastic if stochastic is not None else self.stochastic
        if not self.training:
            stochastic = False
        if stochastic:
            As = self.As + torch.randn_like(self.As) * self.As_noise_scale
            translation = self.translation + torch.randn_like(self.translation) * self.translation_noise_scale
        else:
            As = self.As
            translation = self.translation
        output = torch.einsum('nab, ncb -> nca', As, x) + translation
        return  output


class AffineTransform(nn.Module):
    def __init__(
        self, 
        n_neurons, 
        in_dim=2,
        stochastic=False,
        init_noise_scale=.1,
        allow_scale=True,
        allow_shear=True,
        uniform_scale=False
        ):
        super().__init__()
        self.n_neurons = n_neurons
        self.in_dim = in_dim
        self.stochastic = stochastic
        self.uniform_scale = uniform_scale
        # Parameterize the rotation as angles
        self.angles = nn.Parameter(torch.zeros(n_neurons))  # Rotation angles for each neuron
        
        if allow_scale:
            # Learnable diagonal elements of the scaling matrix
            if uniform_scale: 
                self.scalings = nn.Parameter(torch.ones(n_neurons, 1)) # Only one diagonal elements is learnable
            else:
                self.scalings = nn.Parameter(torch.ones(n_neurons, in_dim)) # Only the diagonal elements are learnable
        else:
            self.register_buffer('scalings', torch.ones(n_neurons, in_dim))

        if allow_shear:
            # Learnable off-diagonal shear elements
            self.shears = nn.Parameter(torch.zeros(n_neurons, in_dim))  # Only learn the shear terms
        else:
            self.register_buffer('shears', torch.zeros(n_neurons, in_dim))

        self.translation = nn.Parameter(torch.zeros(n_neurons, 1, in_dim))

        if stochastic:
            self.init_noise_scale = init_noise_scale
            self._As_noise_scale = nn.Parameter(torch.log(torch.ones(n_neurons, in_dim, in_dim) * init_noise_scale))
            self._translation_noise_scale = nn.Parameter(torch.log(torch.ones(n_neurons, 1, in_dim) * init_noise_scale))
            
    @property
    def As_noise_scale(self):
        return torch.exp(self._As_noise_scale)

    @property
    def translation_noise_scale(self):
        return torch.exp(self._translation_noise_scale)
    
    @property
    def rotation_matrices(self):
        cos_theta = torch.cos(self.angles)
        sin_theta = torch.sin(self.angles)
        return torch.stack([torch.stack([cos_theta, -sin_theta], dim=-1),
                            torch.stack([sin_theta, cos_theta], dim=-1)], dim=-2)
    
    @property
    def scale_matrices(self):
        if self.uniform_scale:
            return torch.diag_embed(self.scalings.repeat(1, 2))
        else:  
            return torch.diag_embed(self.scalings)
    
    @property
    def shear_matrices(self):
        return torch.stack([torch.stack([torch.ones_like(self.shears[:, 0]), self.shears[:, 0]], dim=-1),
                            torch.stack([self.shears[:, 1], torch.ones_like(self.shears[:, 1])], dim=-1)], dim=-2)
    
    @property
    def As(self):
        return torch.einsum('nab,nbc,ncd -> nad', self.rotation_matrices, self.scale_matrices, self.shear_matrices)
        
    def forward(self, x, stochastic=None):
        stochastic = stochastic if stochastic is not None else self.stochastic
        if not self.training:
            stochastic = False
        if stochastic:
            As = self.As + torch.randn_like(self.As) * self.As_noise_scale
            translation = self.translation + torch.randn_like(self.translation) * self.translation_noise_scale
        else:
            As = self.As
            translation = self.translation
        output = torch.einsum('nab, ncb -> nca', As, x) + translation
        return  output



class CoordinateTransform(nn.Module):
    def __init__(
        self,
        n_neurons,
        only_affine=False,
        flow_kwgs=None,
        in_dim=2,
        stochastic=False,
        init_noise_scale=.1,
        allow_scale=True,
        allow_shear=True, 
        uniform_scale=False
        ):
        super().__init__()
        self.n_dim = n_neurons
        self.only_affine = only_affine
        self.stochastic = stochastic
        self.init_noise_scale = init_noise_scale
        self.allow_scale = allow_scale
        self.allow_shear = allow_shear
        self.uniform_scale = uniform_scale
        transforms = []
        if not only_affine:
            if flow_kwgs is None:
                flow_kwgs = dict(n_layers=2, in_dim=2, affine=True, init_identity=True, hidden_features=16, hidden_layers=1)
            else:
                flow_kwgs["in_dim"] = in_dim
            transforms.append(("Warp", RealNVP(n_neurons, **flow_kwgs)))
        transforms.append(("Affine", AffineTransform(
            n_neurons, 
            in_dim=in_dim, 
            stochastic=stochastic, 
            init_noise_scale=init_noise_scale,             
            allow_scale=allow_scale,
            allow_shear=allow_shear,
            uniform_scale=uniform_scale)))
        self.transforms = nn.Sequential(OrderedDict(transforms))
        
    def forward(self, coords):
        if len(coords.shape) == 2:
            coords.unsqueeze_(0)
        return self.transforms(coords)


def cycle(iterable):
    # see https://github.com/pytorch/pytorch/issues/23900
    iterator = iter(iterable)
    while True:
        try:
            yield next(iterator)
        except StopIteration:
            iterator = iter(iterable)


def alternate(*args):
    """
    Given multiple iterators, returns a generator that alternatively visit one element from each iterator at a time.

    Examples:
        >>> list(alternate(['a', 'b', 'c'], [1, 2, 3], ['Mon', 'Tue', 'Wed']))
        ['a', 1, 'Mon', 'b', 2, 'Tue', 'c', 3, 'Wed']

    Args:
        *args: one or more iterables (e.g. tuples, list, iterators) separated by commas

    Returns:
        A generator that alternatively visits one element at a time from the list of iterables
    """
    for row in zip(*args):
        yield from row


def cycle_datasets(loaders):
    """
    Given a dictionary mapping data_key into dataloader objects, returns a generator that alternately yields
    output from the loaders in the dictionary. The order of data_key traversal is determined by the first invocation to `.keys()`.
    To obtain deterministic behavior of key traversal, recommended to use OrderedDict.

    The generator terminates as soon as any one of the constituent loaders is exhausted.

    Args:
        loaders (dict): Dict mapping a data_key to a dataloader object.

    Yields:
        string, Any: data_key  and and the next output from the data loader corresponding to the data_key
    """
    keys = list(loaders.keys())
    # establish a consistent ordering across loaders
    ordered_loaders = [loaders[k] for k in keys]
    for data_key, outputs in zip(cycle(loaders.keys()), alternate(*ordered_loaders)):
        yield data_key, outputs


class Exhauster:
    """
    Given a dictionary of data loaders, mapping data_key into a data loader, steps through each data loader, moving onto the next data loader
    only upon exhausing the content of the current data loader.
    """

    def __init__(self, loaders):
        self.loaders = loaders

    def __iter__(self):
        for data_key, loader in self.loaders.items():
            for batch in loader:
                yield data_key, batch

    def __len__(self):
        return sum([len(loader) for loader in self.loaders])


class LongCycler:
    """
    Cycles through trainloaders until the loader with largest size is exhausted.
        Needed for dataloaders of unequal size (as in the monkey data).
    """

    def __init__(self, loaders):
        self.loaders = loaders
        self.max_batches = max([len(loader) for loader in self.loaders.values()])

    def __iter__(self):
        cycles = [cycle(loader) for loader in self.loaders.values()]
        for k, loader, _ in zip(
            cycle(self.loaders.keys()),
            (cycle(cycles)),
            range(len(self.loaders) * self.max_batches),
        ):
            yield k, next(loader)

    def __len__(self):
        return len(self.loaders) * self.max_batches


class ShortCycler:
    """
    Cycles through trainloaders until the loader with smallest size is exhausted.
        Needed for dataloaders of unequal size (as in the monkey data).
    """

    def __init__(self, loaders):
        self.loaders = loaders
        self.min_batches = min([len(loader) for loader in self.loaders.values()])

    def __iter__(self):
        cycles = [cycle(loader) for loader in self.loaders.values()]
        for k, loader, _ in zip(
            cycle(self.loaders.keys()),
            (cycle(cycles)),
            range(len(self.loaders) * self.min_batches),
        ):
            yield k, next(loader)

    def __len__(self):
        return len(self.loaders) * self.min_batches



from .attention import AttentionReadout
from .base import ClonedReadout, Readout
from .factorized import FullFactorized2d, FullSXF, SpatialXFeatureLinear
from .gaussian import (
    DeterministicGaussian2d,
    FullGaussian2d,
    Gaussian2d,
    Gaussian3d,
    RemappedGaussian2d,
    UltraSparse,
)
from .multi_readout import MultiReadoutBase, MultiReadoutSharedParametersBase
from .point_pooled import PointPooled2d, SpatialTransformerPooled3d
from .pyramid import PointPyramid2d

### Userguide ###

# In order to build your multi-readout, pass the respective readout to your multi-readout base class
# together with your readout kwargs. Use the MultiReadoutSharedParametersBase if you want to share parameters
# between the readouts, otherwise use the MultiReadoutBase. Note that not all readouts support parameter sharing.

# Example:
# standard_multi_pointpooled_readout = MultiReadoutBase(PointPooled2d, **readout_kwargs)
# parameter_sharing_multi_gaussian_readout = MultiReadoutSharedParametersBase(FullGaussian2d, **readout_kwargs)



import torch

from featurevis.utils import varargin


class Ensemble():
    """ Average the response across a set of models.

    Arguments:
        models (list): A list of pytorch models.
        readout_key (str): String identifying the scan whose neurons will be outputted by
            the model
        eye_pos (torch.Tensor): A [1 x 2] tensor with the position of the pupil(x, y).
            This shifts the FOV of all cells. Default (None) is position at center of
            screen (i.e., it disables the shifter).
        behavior (torch.Tensor): A [1 x 3] tensor with the behavior parameters
            (pupil_dilation, dpupil_dilation/dt, treadmill). Default is to return results
            without modulation.
        neuron_idx (int or slice or list): Neuron(s) to return. Default returns all cells.
        average_batch (boolean): If True, responses are averaged across all images in the
            batch (output is a num_neurons tensor). Otherwise, output is a num_images x
            num_neurons tensor.
        device (torch.Device or str): Where to load the models.

    Note:
        We copy the models to avoid overwriting the gradients of the original models. You
        can access our copy of the models as my_ensemble.models.
    """
    def __init__(self, models, readout_key, eye_pos=None, behavior=None,
                 neuron_idx=slice(None), average_batch=True, device='cuda'):
        import copy

        self.models = [copy.deepcopy(m) for m in models]
        self.readout_key = readout_key
        self.eye_pos = None if eye_pos is None else eye_pos.to(device)
        self.behavior = None if behavior is None else behavior.to(device)
        self.neuron_idx = neuron_idx
        self.average_batch = average_batch
        self.device = device

        for m in self.models:
            m.to(device)
            m.eval()

    def __call__(self, x):
        resps = [m(x, self.readout_key, eye_pos=self.eye_pos, behavior=self.behavior)[:,
                 self.neuron_idx] for m in self.models]
        resps = torch.stack(resps)  # num_models x batch_size x num_neurons
        resp = resps.mean(0).mean(0) if self.average_batch else resps.mean(0)

        return resp


class VGG19Core():
    """ A pretrained VGG-19. Output will be intermediate feature representation
    (N x C x H x W) at the desired layer.

    Arguments:
        layer (int): Index (0-based) of the layer that will be optimized.
        use_batchnorm (boolean): Whether to download the version with batchnorm.
        device (torch.Device or str): Where to place the model.
    """
    def __init__(self, layer, use_batchnorm=True, device='cuda'):
        from torchvision import models

        vgg19 = (models.vgg19_bn(pretrained=True) if use_batchnorm else
                 models.vgg19(pretrained=True))
        if layer < len(vgg19.features):
            self.model = vgg19.features[:layer + 1]
        else:
            raise ValueError('layer out of range (max is', len(vgg19.features))
        self.model.to(device)
        self.model.eval()

    @varargin
    def __call__(self, x):
        return self.model(x)


class VGG19():
    """ A pretrained VGG-19. Output will be the average of one channel across spatial
    dimensions.

    Arguments:
        layer (int): Index (0-based) of the layer that will be optimized.
        channel (int)_: Index (0-based) of the channel that will be optimized.
        use_batchnorm (boolean): Whether to download the version with batchnorm.
        device (torch.Device or str): Where to place the model.
    """
    def __init__(self, layer, channel, use_batchnorm=True, device='cuda'):
        self.model = VGG19Core(layer, use_batchnorm, device)
        self.channel = channel

    def __call__(self, x):
        resp = self.model(x)[:, self.channel, :, :].mean()
        return resp



import numpy as np
import torch
from torch import nn
from tqdm import tqdm
from featurevis.ops import GaussianBlur
import matplotlib.pyplot as plt
from lipstick import GifMaker

from .inr import INRTemplates
from .contrastive_loss import SimCLROnGrid
from .utils.training import (
    JitteringGridDatamodule,
    ParamReduceOnPlateau,
    check_activity_requirements,
    ImprovementChecker,
)
from .utils.img_transforms import StandardizeClip, FixEnergyNormClip
from .utils.mask import apply_mask
from .utils.general import SingleNeuronModel, infer_device
# from .utils.mei import check_meis


class InvarianceManifold:
    def __init__(
        self,
        response_predicting_model,
        meis_dict,
        device=None,
        # INR template config
        widths=[50] * 4,
        positional_encoding_dim=50,
        positional_encoding_projection_scale=10,
        aux_positional_encoding_dim=50,
        aux_positional_encoding_projection_scale=0.1,
        periodic_invariance=True,
        nonlinearity=nn.Tanh,
        final_nonlinearity=nn.Tanh,
        weights_scale=0.1,
        bias=True,
        # Optimization config
        required_img_norm=None,
        required_img_std=None,
        pixel_value_lower_bound=None,
        pixel_value_upper_bound=None,
    ):
        # check that input argument are there and valid
        if required_img_std is None and required_img_norm is None:
            raise ValueError("Either std or norm should be provided.")
        elif required_img_std is not None and required_img_norm is not None:
            raise ValueError("Only one of std or norm should be provided.")

        img_normalized_by = "norm" if required_img_norm is not None else "std"

        self.response_predicting_model = response_predicting_model

        input_shape = list(meis_dict.values())[0]["mei"].shape
        channels, height, width = input_shape
        img_res = (height, width)
        template_config = dict(
            img_res=img_res,
            num_templates=1,
            out_channels=channels,
            widths=widths,
            positional_encoding_projection_scale=positional_encoding_projection_scale,
            positional_encoding_dim=positional_encoding_dim,
            aux_positional_encoding_dim=aux_positional_encoding_dim,
            aux_positional_encoding_projection_scale=aux_positional_encoding_projection_scale,
            periodic_invariance=periodic_invariance,
            nonlinearity=nonlinearity,
            final_nonlinearity=final_nonlinearity,
            weights_scale=weights_scale,
            bias=bias,
        )
        self.periodic_invariance = periodic_invariance

        if device is None:
            device = infer_device(response_predicting_model)

        self.template = INRTemplates(**template_config).to(device)

        # check_meis() #TODO: implement this function
        self.meis_dict = meis_dict

        self.pixel_min = pixel_value_lower_bound
        self.pixel_max = pixel_value_upper_bound
        std = required_img_std
        norm = required_img_norm
        mean = (self.pixel_min + self.pixel_max) / 2

        # select img transform
        if img_normalized_by == "std":
            img_transf_config = dict(
                mean=mean, std=std, pixel_min=self.pixel_min, pixel_max=self.pixel_max
            )
            self.img_transforms = StandardizeClip(**img_transf_config).to(device)
        elif img_normalized_by == "norm":
            img_transf_config = dict(
                mean=mean, norm=norm, pixel_min=self.pixel_min, pixel_max=self.pixel_max
            )
            self.img_transforms = FixEnergyNormClip(**img_transf_config).to(device)
        else:
            raise ValueError("img_transform does not match existing names")

    def learn(
        self,
        template_neuron_idx,
        gaussian_blur_sigma=None,
        steps_per_epoch=50,
        grid_points_per_dim=20,
        neighbor_size=0.1,
        temperature=0.3,
        reg_decr=0.8,
        min_epochs=10,
        lr=1e-3,
        additional_epochs=0,
        reg_scale=2,
        requirements=None,
        num_max_epochs=1000,
        verbose=False,
    ):
        device = next(self.template.parameters()).device
        self.template_neuron_idx = template_neuron_idx

        if gaussian_blur_sigma is not None:
            gaussian_blur = GaussianBlur(gaussian_blur_sigma)
        else:
            gaussian_blur = None

        # Dataloader for Training
        dataloader_config = dict(
            num_invariances=1,
            grid_points_per_dim=grid_points_per_dim,
            steps_per_epoch=steps_per_epoch,
        )
        dm = JitteringGridDatamodule(**dataloader_config)
        dm.train_dataloader()
        grid = dm.grid.to(device)

        # Contrastive Regularization Module
        objective_config = dict(
            num_invariances=1,
            grid_points_per_dim=grid_points_per_dim,
            neighbor_size=neighbor_size,
            temperature=temperature,
            with_periodic_invariances=self.periodic_invariance,
            with_round_neighbor=False,
        )
        grid_reg = SimCLROnGrid(**objective_config).to(device)

        # Contrastive Regularization Strength
        reg_scheduler = ParamReduceOnPlateau(
            factor=reg_decr,
            patience=5,
            threshold=0.005,
            mode="max",
        )

        if requirements is None:
            requirements = dict(avg=.99, std=1., necessary_min=0.98)

        optimizer = torch.optim.Adam(self.template.func.parameters(), lr=lr)

        template_neuron_model = SingleNeuronModel(
            self.response_predicting_model, template_neuron_idx
        )

        template_mei_act = self.meis_dict[template_neuron_idx]["activation"]
        template_rf_mask = self.meis_dict[template_neuron_idx]["mask"]

        ignore_scale_scheduler = False
        self.template.train()
        pbar = tqdm(range(num_max_epochs))
        for epoch in pbar:
            for input_grid in dm.train_dataloader():
                input_grid = input_grid.to(device)
                img_pre, img_post, _acts, _ = self.forward(
                    input_grid,
                    self.template,
                    self.img_transforms,
                    template_neuron_model,
                    return_template=True,
                    gb=gaussian_blur,
                )
                acts = _acts / template_mei_act
                within_rf_mask_similarity = grid_reg.reg_term(
                    apply_mask(
                        img_post,
                        torch.from_numpy(template_rf_mask).to(device),
                        self.pixel_min,
                        self.pixel_max,
                    )
                )
                loss = -acts.mean() - reg_scale * within_rf_mask_similarity
                optimizer.zero_grad()
                torch.autograd.set_detect_anomaly(True)
                loss.backward()
                optimizer.step()

            if not ignore_scale_scheduler:
                if epoch > min_epochs:
                    reg_scale = reg_scheduler.step(
                        metric_to_monitor=acts.mean(),
                        value_to_reduce=reg_scale,
                    )
                    with torch.no_grad():
                        img_pre, img_post, _acts, _ = self.forward(
                            grid,
                            self.template,
                            self.img_transforms,
                            template_neuron_model,
                            return_template=True,
                            gb=gaussian_blur,
                        )
                        acts = _acts / template_mei_act
                    if check_activity_requirements(acts, requirements):
                        print("requirements_satisfied")
                        ignore_scale_scheduler = True
                        last_scheduling_epoch = epoch

            if ignore_scale_scheduler:
                if epoch >= (last_scheduling_epoch + additional_epochs):
                    break
            
            act_desc = f"Act mean = {acts.mean().item():.2f} (min = {acts.min().item():.2f} std = {acts.std().item():.2f})"
            desc = act_desc
            if verbose:
                cont_desc = f"Contrastive reg scale = {reg_scale:.3f}"
                improve_desc = f"Epochs without improving = {reg_scheduler.num_epochs_no_improvement} (patience = {reg_scheduler.patience})"
                desc = desc + " | " + cont_desc + " | " + improve_desc

            pbar.set_description(desc)

        self.template.eval()

        with torch.no_grad():
            img_pre, img_post, _acts, _ = self.forward(
                grid,
                self.template,
                self.img_transforms,
                template_neuron_model,
                return_template=True,
                gb=gaussian_blur,
            )
            acts = _acts / template_mei_act

        images_on_template_manifold = img_post.cpu().data.numpy()
        template_neuron_activations = _acts.squeeze().cpu().data.numpy()

        return images_on_template_manifold, template_neuron_activations

    def match(
        self,
        target_neuron_idxs,
        grid_points_per_dim=20,
        steps_per_epoch=1,
        lr=1e-3,
        num_epochs=1000,
        patience=15,
        ignore_diff_smaller_than=1e-3,
        rotate_angle_and_scale=False,
        find_best_translation=False,
        only_affine_coordinate_transformation=True,
        stochastic_coordinate_transformation=False,
        allow_scale_coordinate_transformation=True,
        allow_shear_coordinate_transformation=True,
        uniform_scale_coordinate_transformation=False,
        init_noise_scale_coordinate_transformation=0.1,
        coordinate_transform_clamp_boundaries=True,
        verbose=False,
    ):
        num_target_neurons = len(target_neuron_idxs)
        self.target_neuron_idxs = target_neuron_idxs
        device = next(self.template.parameters()).device

        rf_positions = {neuron_idx: np.array([value["center_pos"]["x"], value["center_pos"]["y"]]).astype(np.float32) for neuron_idx, value in self.meis_dict.items()}
    
        template_rf_mask = self.meis_dict[self.template_neuron_idx]["mask"]
        template_rf_location = rf_positions[self.template_neuron_idx]
        others_rf_mask = np.array([self.meis_dict[idx]["mask"] for idx in target_neuron_idxs])
        others_rf_location = np.array([rf_positions[neuron_idx] for neuron_idx in target_neuron_idxs])
        others_mei_act = np.array([self.meis_dict[idx]["activation"] for idx in target_neuron_idxs])
        others_mei_act = torch.from_numpy(others_mei_act).to(device)
        other_neurons_loc_in_model = target_neuron_idxs
        other_neurons_loc_in_list = np.arange(len(other_neurons_loc_in_model))
       
        # Initialize Coordinate Transformation
        self.template.initialize_coordinate_transform(
            num_neurons=num_target_neurons,
            only_affine=only_affine_coordinate_transformation,
            stochastic=stochastic_coordinate_transformation,
            init_noise_scale=init_noise_scale_coordinate_transformation,
            allow_scale=allow_scale_coordinate_transformation,
            allow_shear=allow_shear_coordinate_transformation,
            uniform_scale=uniform_scale_coordinate_transformation,
            clamp_boundaries=coordinate_transform_clamp_boundaries,
        )

        self.template.register_coords_shifts(
            torch.from_numpy(template_rf_location).to(device), 
            torch.from_numpy(others_rf_location).to(device),
            )

        dataloader_config = dict(
            num_invariances=1,
            grid_points_per_dim=grid_points_per_dim,
            steps_per_epoch=steps_per_epoch,
        )
        grid_dataloader = JitteringGridDatamodule(**dataloader_config)

        template = self.initialize_with_best_angle_and_translation(
            self.template,
            self.response_predicting_model,
            self.img_transforms,
            grid_dataloader,
            other_neurons_loc_in_list,
            other_neurons_loc_in_model,
            template_rf_mask=template_rf_mask,
            others_rf_mask=others_rf_mask,
            rotate_angle_and_scale=rotate_angle_and_scale,
            find_best_translation=find_best_translation,
        )

        optimizer = torch.optim.Adam(template.coordinate_transform.parameters(), lr=lr)
        improvement_checker = ImprovementChecker(
            patience=patience, ignore_diff_smaller_than=ignore_diff_smaller_than
        )
        pbar = tqdm(range(num_epochs))

        # Training Loop
        for epoch in pbar:
            template.train()
            for input_grid in grid_dataloader.train_dataloader():
                input_grid = input_grid.to(device)
                img_pre, img_post, _acts, _ = self.forward(
                    input_grid,
                    template,
                    self.img_transforms,
                    self.response_predicting_model,
                    return_template=False,
                    other_neurons_loc_in_list=other_neurons_loc_in_list,
                    other_neurons_loc_in_model=other_neurons_loc_in_model,
                )
                relevant_acts = _acts[range(len(self.target_neuron_idxs)), :, range(len(self.target_neuron_idxs))]
                acts = relevant_acts.mean(dim=1) / others_mei_act
                loss = -acts.mean()
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

            is_increasing = improvement_checker.is_increasing(acts.mean().item())
            act_desc = f"Activation mean = {acts.mean().item():.2f} (min = {acts.min().item():.2f} max = {acts.max().item():.2f})"
            desc = act_desc
            if verbose:
                improve_desc = f"Epochs without improving = {improvement_checker.has_not_improved_counter} (patience: {improvement_checker.patience})"
                desc = desc + " | " + improve_desc
            pbar.set_description(desc)

            if not is_increasing:
                break

        template.eval()

        with torch.no_grad():
            img_pre, img_post, _acts, _ = self.forward(
                input_grid,
                template,
                self.img_transforms,
                self.response_predicting_model,
                return_template=False,
                other_neurons_loc_in_list=other_neurons_loc_in_list,
                other_neurons_loc_in_model=other_neurons_loc_in_model,
            )
            relevant_acts = _acts[range(len(self.target_neuron_idxs)), :, range(len(self.target_neuron_idxs))]

        images_on_aligned_manifolds = img_post.cpu().data.numpy()
        target_neurons_activations = relevant_acts.cpu().data.numpy()

        return images_on_aligned_manifolds, target_neurons_activations

    def save_learned_template_as_gif(self, n_images=20, name=None, cmap='Greys_r', fig_kws=None):
        fig_kws = fig_kws if fig_kws is not None else {}
        images_from_learned_template = self.get_images_from_learned_template(n_images=n_images)
        name = name if name is not None else f"learned_template_{self.template_neuron_idx}"
        with GifMaker(name) as g:
            for img in images_from_learned_template:
                fig, ax = plt.subplots()
                vmax = np.abs(img).max()
                vmin = -vmax
                ax.imshow(img[0], vmin=vmin, vmax=vmax, cmap=cmap)
                ax.set(xticks=[], yticks=[])
                g.add(fig)
        return g
    
    def save_matched_template_as_gif(self, target_neuron_idx, n_images=20, name=None, cmap='Greys_r', fig_kws=None):
        fig_kws = fig_kws if fig_kws is not None else {}
        images_from_matched_template = self.get_images_from_matched_template(target_neuron_idx, n_images=n_images)
        name = name if name is not None else f"learned_template_{self.template_neuron_idx}_matched_to_{target_neuron_idx}"
        with GifMaker(name) as g:
            for img in images_from_matched_template:
                fig, ax = plt.subplots(**fig_kws)
                vmax = np.abs(img).max()
                vmin = -vmax
                ax.imshow(img[0], vmin=vmin, vmax=vmax, cmap=cmap)
                ax.set(xticks=[], yticks=[])
                g.add(fig)
        return g
    
    def get_images_from_learned_template(self, n_images=20):
        device = next(self.template.parameters()).device
        dataloader_config = dict(
            num_invariances=1,
            grid_points_per_dim=n_images,
            steps_per_epoch=1,
        )
        dm = JitteringGridDatamodule(**dataloader_config)
        dm.train_dataloader()
        grid = dm.grid.to(device)
        with torch.no_grad():
            return self.get_images(
                grid,
                self.template,
                self.img_transforms,
                return_template=True,
            ).cpu().data.numpy()
    
    def get_images_from_matched_template(self, target_neuron_idx, n_images=20):
        if not isinstance(target_neuron_idx, int):
            raise ValueError("target_neuron_idx should be an integer, referring to a single target neuron.")
        if target_neuron_idx not in self.target_neuron_idxs:
            raise ValueError(f"target_neuron_idx should be one of the target neuron indices: {self.target_neuron_idxs}")
        target_neuron_idx_from_zero = np.where(target_neuron_idx == np.array(self.target_neuron_idxs))[0][0]

        device = next(self.template.parameters()).device
        dataloader_config = dict(
            num_invariances=1,
            grid_points_per_dim=n_images,
            steps_per_epoch=1,
        )
        dm = JitteringGridDatamodule(**dataloader_config)
        dm.train_dataloader()
        grid = dm.grid.to(device)

        channels, height, width = self.meis_dict[self.template_neuron_idx]["mei"].shape
        with torch.no_grad():
            images = self.get_images(
                grid,
                self.template,
                self.img_transforms,
                return_template=False,
            ).reshape(len(self.target_neuron_idxs), n_images, channels, height, width)[target_neuron_idx_from_zero]
            return images.cpu().data.numpy()

    def get_images(
            self, 
            grid,
            image_generator,
            img_transf,
            return_template=False,
            centralize_template=False,
            return_pre=False,
            ):
        
        _img_pre = image_generator(
            aux=grid,
            return_template=return_template,
            centralize_template=centralize_template,
        )

        img_pre = _img_pre.flatten(0, -4)
        img_post = img_transf(img_pre)

        if return_pre:
            return _img_pre, img_post
        else: 
            return img_post
        
    def forward(
        self,
        grid,
        image_generator,
        img_transf,
        encoding_model,
        gb=None,
        return_template=False,
        other_neurons_loc_in_list=None,
        other_neurons_loc_in_model=None,
        centralize_template=False,
    ):
        """forward pass throught the pipeline"""
        _img_pre, img_post = self.get_images(
            grid,
            image_generator,
            img_transf,
            return_template=return_template,
            centralize_template=centralize_template,
            return_pre=True,
            )
        
        if return_template:
            *N, C, H, W = _img_pre.shape
        else:
            D, *N, C, H, W = _img_pre.shape

        if gb is not None:
            img_post.register_hook(gb)

        acts = encoding_model(img_post)

        img_post_hres = img_post
        img_pre = _img_pre.flatten(0, -4)
        if return_template:
            return (
                img_pre.reshape(*N, C, H, W),
                img_post.reshape(*N, C, H, W),
                acts.reshape(*N, -1),
                img_post_hres.reshape(*N, C, H, W),
            )
        else:
            D_selected = len(other_neurons_loc_in_model)
            return (
                img_pre.reshape(D, *N, C, H, W)[other_neurons_loc_in_list],
                img_post.reshape(D, *N, C, H, W)[other_neurons_loc_in_list],
                acts[:, other_neurons_loc_in_model].reshape(D, *N, D_selected)[
                    other_neurons_loc_in_list
                ],
                img_post_hres.reshape(D, *N, C, H, W)[other_neurons_loc_in_list],
            )

    def initialize_with_best_angle_and_translation(
        self,
        template, 
        encoding_model, 
        img_transforms,
        grid_dataloader,  
        other_neurons_loc_in_list, 
        other_neurons_loc_in_model, 
        template_rf_mask=None, 
        others_rf_mask=None, 
        rotate_angle_and_scale=False,
        find_best_translation=False
    ):
        device = next(template.parameters()).device
        if rotate_angle_and_scale:
            if others_rf_mask is None:
                raise ValueError("others_rf_mask must be provided if rotate_angle_and_scale is True")
            # Initialize the transformation at best rotation angle (and scale) for each target neuron
            mask_size_fraction = others_rf_mask.sum(axis=(1, 2)) / template_rf_mask.sum()
            mask_size_fraction = torch.from_numpy(mask_size_fraction).to(device)

        angles = np.linspace(0, 2*np.pi, int(360/6 + 1))[:-1].astype(np.float32)

        if find_best_translation:
            translations = np.linspace(-.1, .1, 11).astype(np.float32)  # Define a range for translations
            desc = 'Finding best angle and translation for initalization'
        else:
            translations = np.array([0.0]).astype(np.float32)
            desc = 'Finding best angle for initalization'

        # Initialize a 4D array to store activations
        activations_for_transformations = np.zeros((len(angles), len(translations), len(translations), len(other_neurons_loc_in_list)))

        # Try different angles and translations to find the best combination for initialization
        with torch.no_grad():
            input_grid = grid_dataloader.grid.to(device)
            img_posts = []
            for i, angle in enumerate(tqdm(angles, desc=desc)):
                for j, tx in enumerate(translations):
                    for k, ty in enumerate(translations):
                        template.coordinate_transform.transforms.Affine.angles.data[other_neurons_loc_in_list] =  torch.ones_like(template.coordinate_transform.transforms.Affine.angles.data[other_neurons_loc_in_list]) * angle
                        if rotate_angle_and_scale:
                            template.coordinate_transform.transforms.Affine.scalings.data[other_neurons_loc_in_list] = torch.ones_like(template.coordinate_transform.transforms.Affine.scalings.data[other_neurons_loc_in_list]) * 1/mask_size_fraction.reshape(-1, 1)

                        template.coordinate_transform.transforms.Affine.translation.data[other_neurons_loc_in_list, 0, 0] = torch.ones_like(template.coordinate_transform.transforms.Affine.translation.data[other_neurons_loc_in_list, 0, 0]) * tx
                        template.coordinate_transform.transforms.Affine.translation.data[other_neurons_loc_in_list, 0, 1] = torch.ones_like(template.coordinate_transform.transforms.Affine.translation.data[other_neurons_loc_in_list, 0, 1]) * ty

                        img_pre, img_post, _acts, _ = self.forward(input_grid, template, img_transforms, encoding_model, return_template=False, 
                                                            other_neurons_loc_in_list=other_neurons_loc_in_list, other_neurons_loc_in_model=other_neurons_loc_in_model)
                        # relevant_acts = torch.diag(_acts.mean(dim=1)).cpu().data.numpy()  # Get activations for each neuron
                        relevant_acts = _acts[range(len(self.target_neuron_idxs)), :, range(len(self.target_neuron_idxs))]
                        relevant_acts = relevant_acts.mean(dim=1).cpu().data.numpy()

                        # Store activations in the 4D array
                        activations_for_transformations[i, j, k, :] = relevant_acts
                        img_posts.append(img_post.cpu().data.numpy())

        # Find the combination of angle, tx, and ty that maximizes the activation for each neuron
        max_indices = np.argmax(activations_for_transformations.reshape(-1, activations_for_transformations.shape[-1]), axis=0)
        best_angle_indices, best_tx_indices, best_ty_indices = np.unravel_index(max_indices, activations_for_transformations.shape[:3])
        best_angles = angles[best_angle_indices]  # Best angles for each neuron
        best_txs = translations[best_tx_indices]  # Best tx for each neuron
        best_tys = translations[best_ty_indices]  # Best ty for each neuron

        # Set the optimal angle and translation for each neuron
        template.coordinate_transform.transforms.Affine.angles.data[other_neurons_loc_in_list] = torch.from_numpy(best_angles).to(device)
        template.coordinate_transform.transforms.Affine.translation.data[other_neurons_loc_in_list, 0, 0] = torch.from_numpy(best_txs).to(device)
        template.coordinate_transform.transforms.Affine.translation.data[other_neurons_loc_in_list, 0, 1] = torch.from_numpy(best_tys).to(device)

        if rotate_angle_and_scale:
            template.coordinate_transform.transforms.Affine.scalings.data[other_neurons_loc_in_list] = torch.ones_like(template.coordinate_transform.transforms.Affine.scalings.data[other_neurons_loc_in_list]) * 1/mask_size_fraction.reshape(-1, 1)
            
        return template



import numpy as np
import torch
from torch import nn as nn

from .base import Readout


class FullFactorized2d(Readout):
    """
    Factorized fully connected layer. Weights are a sum of outer products between a spatial filter and a feature vector.
    """

    def __init__(
        self,
        in_shape,
        outdims,
        bias,
        normalize=True,
        init_noise=1e-3,
        constrain_pos=False,
        positive_weights=False,
        shared_features=None,
        mean_activity=None,
        spatial_and_feature_reg_weight=1.0,
        gamma_readout=None,  # depricated, use feature_reg_weight instead
        **kwargs,
    ):

        super().__init__()

        c, w, h = in_shape
        self.in_shape = in_shape
        self.outdims = outdims
        self.positive_weights = positive_weights
        self.constrain_pos = constrain_pos
        self.init_noise = init_noise
        self.normalize = normalize
        self.mean_activity = mean_activity
        self.spatial_and_feature_reg_weight = self.resolve_deprecated_gamma_readout(
            spatial_and_feature_reg_weight, gamma_readout
        )

        self._original_features = True
        self.initialize_features(**(shared_features or {}))
        self.spatial = nn.Parameter(torch.Tensor(self.outdims, w, h))

        if bias:
            bias = nn.Parameter(torch.Tensor(outdims))
            self.register_parameter("bias", bias)
        else:
            self.register_parameter("bias", None)

        self.initialize(mean_activity)

    @property
    def shared_features(self):
        return self._features

    @property
    def features(self):
        if self._shared_features:
            return self.scales * self._features[self.feature_sharing_index, ...]
        else:
            return self._features

    @property
    def weight(self):
        if self.positive_weights:
            self.features.data.clamp_min_(0)
        n = self.outdims
        c, w, h = self.in_shape
        return self.normalized_spatial.view(n, 1, w, h) * self.features.view(n, c, 1, 1)

    @property
    def normalized_spatial(self):
        """
        Normalize the spatial mask
        """
        if self.normalize:
            norm = self.spatial.pow(2).sum(dim=1, keepdim=True)
            norm = norm.sum(dim=2, keepdim=True).sqrt().expand_as(self.spatial) + 1e-6
            weight = self.spatial / norm
        else:
            weight = self.spatial
        if self.constrain_pos:
            weight.data.clamp_min_(0)
        return weight

    def regularizer(self, reduction="sum", average=None):
        return self.l1(reduction=reduction, average=average) * self.spatial_and_feature_reg_weight

    def l1(self, reduction="sum", average=None):
        reduction = self.resolve_reduction_method(reduction=reduction, average=average)
        if reduction is None:
            raise ValueError("Reduction of None is not supported in this regularizer")

        n = self.outdims
        c, w, h = self.in_shape
        ret = (
            self.normalized_spatial.view(self.outdims, -1).abs().sum(dim=1, keepdim=True)
            * self.features.view(self.outdims, -1).abs().sum(dim=1)
        ).sum()
        if reduction == "mean":
            ret = ret / (n * c * w * h)
        return ret

    def initialize(self, mean_activity=None):
        """
        Initializes the mean, and sigma of the Gaussian readout along with the features weights
        """
        if mean_activity is None:
            mean_activity = self.mean_activity
        self.spatial.data.normal_(0, self.init_noise)
        self._features.data.normal_(0, self.init_noise)
        if self._shared_features:
            self.scales.data.fill_(1.0)
        if self.bias is not None:
            self.initialize_bias(mean_activity=mean_activity)

    def initialize_features(self, match_ids=None, shared_features=None):
        """
        The internal attribute `_original_features` in this function denotes whether this instance of the FullGuassian2d
        learns the original features (True) or if it uses a copy of the features from another instance of FullGaussian2d
        via the `shared_features` (False). If it uses a copy, the feature_l1 regularizer for this copy will return 0
        """
        c, w, h = self.in_shape
        if match_ids is not None:
            assert self.outdims == len(match_ids)

            n_match_ids = len(np.unique(match_ids))
            if shared_features is not None:
                assert shared_features.shape == (
                    n_match_ids,
                    c,
                ), f"shared features need to have shape ({n_match_ids}, {c})"
                self._features = shared_features
                self._original_features = False
            else:
                self._features = nn.Parameter(
                    torch.Tensor(n_match_ids, c)
                )  # feature weights for each channel of the core
            self.scales = nn.Parameter(torch.Tensor(self.outdims, 1))  # feature weights for each channel of the core
            _, sharing_idx = np.unique(match_ids, return_inverse=True)
            self.register_buffer("feature_sharing_index", torch.from_numpy(sharing_idx))
            self._shared_features = True
        else:
            self._features = nn.Parameter(torch.Tensor(self.outdims, c))  # feature weights for each channel of the core
            self._shared_features = False

    def forward(self, x, shift=None, **kwargs):
        if shift is not None:
            raise NotImplementedError("shift is not implemented for this readout")
        if self.constrain_pos:
            self.features.data.clamp_min_(0)

        N, c, w, h = x.size()
        c_in, w_in, h_in = self.in_shape
        if (c_in, w_in, h_in) != (c, w, h):
            raise ValueError("the specified feature map dimension is not the readout's expected input dimension")

        y = torch.einsum("ncwh,owh->nco", x, self.normalized_spatial)
        y = torch.einsum("nco,oc->no", y, self.features)
        if self.bias is not None:
            y = y + self.bias
        return y

    def __repr__(self):
        c, w, h = self.in_shape
        r = self.__class__.__name__ + " (" + "{} x {} x {}".format(c, w, h) + " -> " + str(self.outdims) + ")"
        if self.bias is not None:
            r += " with bias"
        if self._shared_features:
            r += ", with {} features".format("original" if self._original_features else "shared")
        if self.normalize:
            r += ", normalized"
        else:
            r += ", unnormalized"
        for ch in self.children():
            r += "  -> " + ch.__repr__() + "\n"
        return r


# Classes for backwards compatibility
class SpatialXFeatureLinear(FullFactorized2d):
    pass


class FullSXF(FullFactorized2d):
    pass



"""
This module implements tracker classes that can be used to keep track
of statistics that are generated throughout training.
"""

from __future__ import annotations

import copy
import logging
import time
from collections import abc, defaultdict
from typing import Dict, List, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
from tqdm import tqdm

from .utils import deep_update

logger = logging.getLogger(__name__)


class Tracker:
    """
    Abstract Tracker class to serve as the bass class for all trackers.
    Defines the two interfacing methods of `log_objective` and `finalize`.

    """

    def log_objective(self, obj=None):
        """
        Logs the provided object

        Args:
            obj (Any, optional): Object to be logged

        Raises:
            NotImplementedError: Override this method to provide a functional behavior.
        """
        raise NotImplementedError("Please override this method to provide functional behavior")

    def finalize(self, obj):
        pass


class TimeObjectiveTracker(Tracker):
    """
    Provides basic tracking of any object with a timestamp. Invoking `finalize()` will
    make all recorded timestamps relative to the first event tracked unless a specific
    time value to relativize against is provided.
    """

    def __init__(self, add_creation_event=False):
        """
        Initializes the tracker. If `add_creation_event` is True, then an entry is created with value
        `0.0` with the timestamp corresponding to the creation fo this tracker object.

        Args:
            add_creation_event (bool, optional): If set to True, an event for creation with value of 0.0 is added to the log. Defaults to False.
        """
        self.tracker = np.array([[time.time(), 0.0]]) if add_creation_event else np.empty((0, 0))

    def log_objective(self, obj):
        """
        Logs the provided object paired with the timestamp. Before finalizing by invoking `finalize()`, all events
        are logged with absolute time in epoch time in seconds.

        Args:
            obj (Any): Object to be logged.
        """
        new_track_point = np.array([[time.time(), obj]])
        self.tracker = np.concatenate((self.tracker, new_track_point), axis=0)

    def finalize(self, reference=None):
        """
        When invoked, all logs time entries are relativized against the first log entry time.
        Pass value `reference` to set the time relative to the passed-in value instead.

        Args:
            reference (float, optional): Timestamp to relativize all logged even times to. If None, relativizes all
                time entries to first time entry of the log. Defaults to None.
        """
        # relativize to the first entry of the tracked events unless reference provided
        reference = self.tracker[0, 0] if reference is None else reference
        # relativize the entry
        self.tracker[:, 0] -= reference


class MultipleObjectiveTracker(Tracker):
    """
    Given a dictionary of functions, this will invoke all functions and log the returned values against
    the invocation timestamp. Calling `finalize` relativizes the timestamps to the first entry (i.e. first
    log_objective call) made.
    """

    def __init__(self, default_name=None, **objectives):
        """
        Initializes the tracker. Pass any additional objective functions as keywords arguments.

        Args:
            default_name (string, optional): Name under which the objective value passed into `log_objective` is saved under.
                If set to None, the passed in value is NOT saved. Defaults to None.
        """
        self._default_name = default_name
        self.objectives = objectives
        self.log = defaultdict(list)
        self.time = []

    def log_objective(self, obj=None):
        """
        Log the objective and also returned values of any list of objective functions this tracker
        is configured with. The passed in value of `obj` is logged only if `default_name` was set to
        something except for None.

        Args:
            obj (Any, optional): Value to be logged if `default_name` is not None. Defaults to None.
        """
        t = time.time()
        # iterate through and invoke each objective function and accumulate the
        # returns. This is performed separately from actual logging so that any
        # error in one of the objective evaluation will halt the whole timestamp
        # logging, and avoid leaving partial results
        values = {}
        # log the passed in value only if `default_name` was given
        if self._default_name is not None:
            values[self._default_name] = obj

        for name, objective in self.objectives.items():
            values[name] = objective()

        # Actually log all values
        self.time.append(t)
        for name, value in values.items():
            self.log[name].append(value)

    def finalize(self, reference=None):
        """
        When invoked, all logs are convereted into numpy arrays and are relativized against the first log entry time.
        Pass value `reference` to set the time relative to the passed-in value instead.

        Args:
            reference (float, optional): Timestamp to relativize all logged even times to. If None, relativizes all
                time entries to first time entry of the log. Defaults to None.
        """
        self.time = np.array(self.time)
        reference = self.time[0] if reference is None else reference
        self.time -= reference
        for k, l in self.log.items():
            self.log[k] = np.array(l)

    def asdict(self, time_key="time", make_copy=True):
        """
        Output the cotent of the tracker as a single dictionary. The time value


        Args:
            time_key (str, optional): Name of the key to save the time information as. Defaults to "time".
            make_copy (bool, optional): If True, the returned log times will be a (shallow) copy. Defaults to True.

        Returns:
            dict: Dictionary containing tracked values as well as the time
        """
        log_copy = {k: np.copy(v) if make_copy else v for k, v in self.log.items()}
        log_copy[time_key] = np.copy(self.time) if make_copy else self.time
        return log_copy


class AdvancedTracker(Tracker):
    """
    This class implements a more advanced, universal tracker that offers many useful features:
     - Store logging information in an arbitrary hierarchical structure, e.g.:
       ```
       {
        "learning_rate" : [0.1, 0.08],
        "Training":
        {
            "img_classification": {
               "accuracy" : [345.234234, 242.34],
               "loss" : [0.342342, 0.78432],
               "normalization": [4,3],
            }
            "neural_prediction": {
               "accuracy" : [345.234234, 242.34],
               "loss" : [0.342342, 0.78432],
               "normalization": [4,3],
            }
        }
        "Validation": {
             "img_classification": {
               "accuracy" : [348.234234, 0.0],
               "loss" : [0.42342, 0.0],
               "normalization": [2,1],
             },
             "patience": [0],
         },
       }
       ```
     - Retrieving and manipulating information via a simple hierarchical key,
       e.g. `("Training","img_classification","loss")`
     - Scores for the same epoch are automatically accumulated via `log_objective(...)` (usually unnormalized)
     - Scores can be automatically normalized if `"normalization"` (e.g. total examples) is tracked on the same hierarchy level
     - Automatic per-epoch initialization for each objective when a new epoch is started via `start_epoch()`
     - Automatic output of current logs to a logger or even tqdm
     - Easy extraction and reloading of tracker state via `state_dict()`, `from_dict(sd)` and `load_state_dict(sd)`


     Example usage:
     1. Specify the log structure and initial values:
         ```
        objectives = {
            "lr": 0,
            "training": {
                "img_classification": {"loss": 0, "accuracy": 0, "normalization": 0}
            },
            "validation": {
                "img_classification": {
                    "loss": 0,
                    "accuracy": 0,
                    "normalization": 0,
                },
                "patience": 0,
            },
        }
        tracker = AdvancedTracker(
            main_objective=("img_classification", "accuracy"), **objectives
        )
         ```
     2. Start a new epoch:
     ```
        self.tracker.start_epoch()
     ```
     3. Log the objectives:
     ```
        tracker.log_objective(
                100 * predicted.eq(targets).sum().item(),
                keys=(mode, task_key, "accuracy"),
            )
        tracker.log_objective(
            batch_size,
            keys=(mode, task_key, "normalization"),
        )
        tracker.log_objective(
            loss.item() * batch_size,
            keys=(mode, task_key, "loss"),
        )
    ```
    4. Display the current objective values, e.g. everything related to training:
    ```
        self.tracker.display_log(tqdm_iterator=t, keys=("training",))
    ```
    5. Save the tracker (i.e. save the training progress):
    ```
        self.tracker.state_dict()
    ```
    """

    def __init__(self, main_objective: Tuple[str, ...] = (), **objectives: Mapping) -> None:
        """
        In principle, `objectives` is expected to be a dictionary of dictionaries
        The hierarchy can in principle be arbitrarily deep.
        The only restriction is that the lowest level has to be a dictionary with values being
        either a numerical value which will be interpreted as the initial value for this objective
        (to be accumulated manually) or a callable (e.g. a function) that returns the objective value.

        Args:
            main_objective: key of the main objective that is used to e.g. decide lr reductions
                            can be something like `("img_classification","accuracy")` to always look at accuracy.
                            Can be combined with a setting specific key in `get_current_main_objective(...)`
            **objectives: e.g. {"dataset": {"objective1": o_fct1, "objective2": 0, "normalization": 0},...}
                           or {"dataset": {"task_key": {"objective1": o_fct1, "objective2": 0},...},...}.
                          Here the key "normalization" is optional on each hierarchy level.
                          If "normalization" exists, then this entry is expected to contain the value that
                          is used to normalize all other values on the same level,
                          and the normalization will be applied whenever the log is returned to the outside.
                          E.g. if a loss is supposed to be tracked, then the "loss" entry will contain the
                          un-normalized accumulation of loss for all inputs in that epoch at any point.
                          To get the normalized loss that is commonly used in practice, simply accumulate the
                          total number of inputs in the "normalization" entry and let AdvancedTracker do the rest.


        """
        self.objectives = objectives
        self.log = self._initialize_log(objectives)
        self.time = []
        self.main_objective = main_objective
        self.epoch = -1  # will be increased to 0 in next line
        self.start_epoch()

    def add_objectives(self, objectives: Mapping, init_epoch: bool = False) -> None:
        """
        Add new objectives (with initial values) to the logs.
        Args:
            objectives: dictionary that needs to follow the same structure as self.log
            init_epoch: flag that decides whether the a new epoch is initialized for this objective
        """
        deep_update(self.objectives, objectives)
        new_log = self._initialize_log(objectives)
        if init_epoch:
            self._initialize_epoch(new_log, objectives)
        deep_update(self.log, new_log)

    def start_epoch(self, append_epoch: bool = True) -> None:
        """Start a new epoch. Initialize each accumulation with its default value."""
        t = time.time()
        self.time.append(t)
        if append_epoch:
            self.epoch += 1
        else:
            self.epoch = 0
        self._initialize_epoch(self.log, self.objectives, append_epoch=append_epoch)

    def log_objective(self, value: float, key: Tuple[str, ...] = ()) -> None:
        """
        Add a new entry to the logs
        Args:
            value: objective score
            key: hierarchical key to match in `self.log`
        """
        if key:
            self._log_objective_value(value, self.log, key)
        else:
            self._log_objective_callables(self.log, self.objectives)

    def display_log(self, key: Tuple[str, ...] = (), tqdm_iterator: Optional[tqdm] = None) -> None:
        """
        Display the current objective value of everything under `key`
        Args:
            key: Tuple describing the objective to display.
                 This could be something like `("Training","img_classification","loss")`
                 to display the current classification loss or something like `("Training","img_classification")`
                 to display everything we save for image classifcation training.
            tqdm_iterator: A tqdm object that the log could be displayed on.
        """
        # normalize (if applicable) and turn into np.arrays:
        n_log = self._normalize_log(self.log)
        # flatten a subset of the dictionary:
        current_log = self._gather_log(n_log, key, index=-1)
        if tqdm_iterator:
            tqdm_iterator.set_postfix(**current_log)
        else:
            logger.info(current_log)

    def get_objective(self, log: Optional[Union[Mapping, Sequence]] = None, key: Tuple[str, ...] = ()) -> np.array:
        """
        Get the value of the objective that corresponds to the key.
        Args:
            log: log to retrieve objective from
            key: key to match with the log

        Returns:
            value array (across epochs and normalized) for a specific objective key
        """
        if log is None:
            log = self._normalize_log(self.log)
        if len(key) > 1:
            return self.get_objective(log[key[0]], key[1:])
        else:
            if isinstance(log[key[0]], (list, np.ndarray)):
                return log[key[0]]
            else:
                raise ValueError("The key does not fully match an objective. Try specifying the complete key sequence.")

    def get_current_objective(self, key: Tuple[str, ...]) -> float:
        return self.get_objective(self._normalize_log(self.log), key)[-1]

    def get_current_main_objective(self, key: Tuple[str, ...]) -> float:
        """
        Main objective is saved in tracker to make it convenient to get the main objective
        that would be used e.g. for learning-rate reduction or similar.
        """
        combined_key = key + self.main_objective if isinstance(key, tuple) else (key,) + self.main_objective
        return self.get_current_objective(combined_key)

    def check_isfinite(self, log: Optional[Union[Mapping, Sequence]] = None) -> bool:
        """
        Checks if all entries in `log` or (normalized) `self.log` are finite.
        Args:
            log: dict that is recursively searched for infinite entries

        Returns: True if all entries are finite.
        """
        if log is None:
            log = self._normalize_log(self.log)
        if isinstance(log, abc.Mapping):
            for k, l in log.items():
                if not self._check_isfinite(l):
                    return False
        else:
            return np.isfinite(log).any()
        return True

    def finalize(self) -> None:
        """After training, normalize the log and save the total time."""
        self.time = np.array(self.time)
        self.time -= self.time[0]
        self.log = self._normalize_log(self.log)

    def state_dict(self) -> Dict[str, any]:
        """
        Serializes this instance to a Python dictionary.

        Returns:
            Dict[str, any]: All attributes that make up this configuration instance
        """
        return copy.deepcopy(self.__dict__)

    def load_state_dict(self, tracker_dict: Dict) -> None:
        """
        Loads given state_dict from another tracker.
        Args:
            tracker_dict: state that should override this tracker

        """
        self.main_objective = tracker_dict["main_objective"]
        self.objectives = tracker_dict["objectives"]
        self.log = tracker_dict["log"]
        self.time = tracker_dict["time"]
        self.epoch = tracker_dict["epoch"]

    @classmethod
    def from_dict(cls, tracker_dict: Dict) -> AdvancedTracker:
        """
        Same as `load_state_dict`, but creates a new tracker from scratch.
        Args:
            tracker_dict: state of tracker that should be loaded in new tracker

        Returns:
            new tracker with `tracker_dict` loaded

        """
        tracker = cls(main_objective=tracker_dict["main_objective"], **tracker_dict["objectives"])
        tracker.log = tracker_dict["log"]
        tracker.time = tracker_dict["time"]
        tracker.epoch = tracker_dict["epoch"]
        return tracker

    def _initialize_log(self, objectives: Mapping[str, any]) -> Dict:
        log = {}
        for key, objective in objectives.items():
            if isinstance(objective, abc.Mapping):
                log[key] = self._initialize_log(objective)
            elif not callable(objective):
                log[key] = []
        return log

    def _initialize_epoch(
        self, log: Union[Mapping, Sequence], objectives: Mapping[str, any], append_epoch: bool = True
    ) -> None:
        """
        For each key in `objectives`, go through log and append a new entry to its list.
        The entry reflects the default value saved in `objectives`.
        Args:
            log: sub dictionary to add the objectives to
            objectives: dictionary of default values
        """
        for key, objective in objectives.items():
            if isinstance(objective, abc.Mapping):
                self._initialize_epoch(log[key], objective)
            elif not callable(objective):
                if append_epoch or len(log[key]) == 0:
                    while len(log[key]) <= self.epoch:
                        log[key].append(objective)
                else:
                    log[key][0] = objective

    def _log_objective_value(self, value: float, log: Union[Mapping, Sequence], key: Tuple[str, ...] = ()) -> None:
        """
        Recursively walk through the log dictionary to get to follow `key`.
        When on lowest level: add `value` to entry at current epoch.
        Args:
            value: objective value to log
            log: log subdict to add value to
            key: key for where `value` is saved
        """
        if len(key) > 1:
            self._log_objective_value(value, log[key[0]], key[1:])
        else:
            log[key[0]][self.epoch] += value

    def _log_objective_callables(self, log: Union[Mapping, Sequence], objectives: Mapping[str, any]) -> None:
        """
        Log all objectives that are specified as callables
        Disclaimer: this is not very well tested!
        Args:
            log:
            objectives:
        """
        for key, objective in objectives.items():
            if isinstance(objective, abc.Mapping):
                self._log_objective_callables(log[key], objective)
            elif callable(objective):
                log[key][self.epoch] += objective()

    def _normalize_log(self, log: Union[Mapping, Sequence]) -> Union[Mapping, np.array]:
        """
        Recursively go through the log and normalize the entries that
        have `"normalization"` information on the same level.
        Args:
            log: subdict on which to apply normalization or list (if on lowest level)

        Returns:
            Normalized log dictionary that has numpy arrays on the lowest level

        """
        if isinstance(log, abc.Mapping):
            n_log = {}
            norm = None
            for key, l in log.items():
                res = self._normalize_log(l)  # to turn into arrays
                if key == "normalization":
                    assert isinstance(res, np.ndarray)
                    norm = res
                else:
                    n_log[key] = res
            self._normalize(n_log, norm)
            return n_log
        else:
            return np.array(log)

    def _normalize(self, n_log: Mapping, norm: np.ndarray):
        """
        Normalizes sub-dict `n_log`
        """
        if norm is not None:
            nonzero_start = (norm != 0).argmax(axis=0)
            norm = norm[nonzero_start:]
            for key, l in n_log.items():
                l = l[nonzero_start:]
                if isinstance(l, np.ndarray):
                    n_log[key] = l / np.where(norm > 0, norm, np.ones_like(norm))

    def _gather_log(self, log: Union[Mapping, Sequence], key: Tuple[str, ...], index: int = -1) -> Dict[str, str]:
        """
        Get a flattened and print-ready version of the log dictionary for a given key
        Args:
            log: subdict to retrieve values from
            key: tuple describing on which level to retrieve from
            index: which epoch to retrieve from

        Returns:
            Flattened dictionary, e.g. `{"img_clasisification accuracy": 98.5, "img_classsifcation loss": 0.456}
        """
        if len(key) > 1:
            return self._gather_log(log[key[0]], key[1:], index)
        elif key:
            return self._gather_log(log[key[0]], (), index)
        elif isinstance(log, abc.Mapping):
            gathered = {}
            for key, l in log.items():
                logs = self._gather_log(l, (), index)
                for k, v in logs.items():
                    gathered[key + " " + k] = v
            return gathered
        else:
            return {"": "{:03.5f}".format(log[index])}



import copy
import logging
from collections import OrderedDict

import numpy as np
import torch

logger = logging.getLogger(__name__)


def copy_state(model):
    """
    Given PyTorch module `model`, makes a copy of the state onto CPU.
    Args:
        model: PyTorch module to copy state dict of

    Returns:
        A copy of state dict with all tensors allocated on the CPU
    """
    copy_dict = OrderedDict()
    state_dict = model.state_dict()
    for k, v in state_dict.items():
        if torch.is_tensor(v):
            copy_dict[k] = v.cpu() if v.is_cuda else v.clone()
        else:
            copy_dict[k] = copy.deepcopy(v)

    return copy_dict


def early_stopping(
    model,
    objective,
    interval=5,
    patience=20,
    start=0,
    max_iter=1000,
    maximize=True,
    tolerance=1e-5,
    switch_mode=True,
    restore_best=True,
    tracker=None,
    scheduler=None,
    lr_decay_steps=1,
):

    """
    Early stopping iterator. Keeps track of the best model state during training. Resets the model to its
        best state, when either the number of maximum epochs or the patience [number of epochs without improvement)
        is reached.
    Also includes a convenient way to reduce the learning rate. Takes as an additional input a PyTorch scheduler object
        (e.g. torch.optim.lr_scheduler.ReduceLROnPlateau), which will automatically decrease the learning rate.
        If the patience counter is reached, the scheduler will decay the LR, and the model is set back to its best state.
        This loop will continue for n times in the variable lr_decay_steps. The patience and tolerance parameters in
        early stopping and the scheduler object should be identical.


    Args:
        model:     model that is being optimized
        objective: objective function that is used for early stopping. The function must accept single positional argument `model`
            and return a single scalar quantity.
        interval:  interval at which objective is evaluated to consider early stopping
        patience:  number of continuous epochs the objective could remain without improvement before the iterator terminates
        start:     start value for iteration (used to check against `max_iter`)
        max_iter:  maximum number of iterations before the iterator terminated
        maximize:  whether the objective is maximized of minimized
        tolerance: margin by which the new objective score must improve to be considered as an update in best score
        switch_mode: whether to switch model's train mode into eval prior to objective evaluation. If True (default),
                     the model is switched to eval mode before objective evaluation and restored to its previous mode
                     after the evaluation.
        restore_best: whether to restore the best scoring model state at the end of early stopping
        tracker (Tracker):
            Tracker to be invoked for every epoch. `log_objective` is invoked with the current value of `objective`. Note that `finalize`
            method is NOT invoked.
        scheduler:  scheduler object, which automatically reduces decreases the LR by a specified amount.
                    The scheduler's `step` method is invoked, passing in the current value of `objective`
        lr_decay_steps: Number of times the learning rate should be reduced before stopping the training.

    """
    training_status = model.training

    def _objective():
        if switch_mode:
            model.eval()
        ret = objective(model)
        if switch_mode:
            model.train(training_status)
        return ret

    def decay_lr(model, best_state_dict):
        old_objective = _objective()
        if restore_best:
            model.load_state_dict(best_state_dict)
            logger.info(f"Restoring best model after lr decay! {old_objective:.6f} ---> {_objective():.6f}")

    def finalize(model, best_state_dict):
        old_objective = _objective()
        if restore_best:
            model.load_state_dict(best_state_dict)
            logger.info(f"Restoring best model! {old_objective:.6f} ---> {_objective():.6f}")
        else:
            logger.info(f"Final best model! objective {_objective():.6f}")

    epoch = start
    # turn into a sign
    maximize = -1 if maximize else 1
    best_objective = current_objective = _objective()
    best_state_dict = copy_state(model)

    for repeat in range(lr_decay_steps):
        patience_counter = 0

        while patience_counter < patience and epoch < max_iter:

            for _ in range(interval):
                epoch += 1
                if tracker is not None:
                    tracker.log_objective(current_objective)
                if (~np.isfinite(current_objective)).any():
                    logger.warning("Objective is not Finite. Stopping training")
                    finalize(model, best_state_dict)
                    return
                yield epoch, current_objective

            current_objective = _objective()

            # if a scheduler is defined, a .step with the current objective is all that is needed to reduce the LR
            if scheduler is not None:
                scheduler.step(current_objective)

            if current_objective * maximize < best_objective * maximize - tolerance:
                logger.info(f"[{epoch:03d}|{patience_counter:02d}/{patience:02d}] ---> {current_objective}")
                best_state_dict = copy_state(model)
                best_objective = current_objective
                patience_counter = 0
            else:
                patience_counter += 1
                logger.info(f"[{epoch:03d}|{patience_counter:02d}/{patience:02d}] ---> {current_objective}")

        if (epoch < max_iter) & (lr_decay_steps > 1) & (repeat < lr_decay_steps):
            decay_lr(model, best_state_dict)

    finalize(model, best_state_dict)



import logging
import warnings

import torch
from torch import nn
from torch.nn import ModuleDict

from .base import Shifter

logger = logging.getLogger(__name__)


class StaticAffine2d(nn.Linear):
    def __init__(self, input_channels=2, output_channels=2, bias=True):
        """
        A simple FC network with bias between input and output channels without a hidden layer.
        Args:
            input_channels (int): number of input channels.
            output_channels (int): number of output channels.
            bias (bool): Adds a bias parameter if True.
        """
        super().__init__(input_channels, output_channels, bias=bias)

    def forward(self, x, trial_idx=None):
        if trial_idx is not None:
            warnings.warn(
                "Trial index was passed but is not used because this shifter network does not support trial indexing."
            )
        x = super().forward(x)
        return torch.tanh(x)

    def initialize(self, bias=None):
        self.weight.data.normal_(0, 1e-6)
        if self.bias is not None:
            if bias is not None:
                logger.info("Setting bias to predefined value")
                self.bias.data = bias
            else:
                self.bias.data.normal_(0, 1e-6)

    def regularizer(self):
        return self.weight.pow(2).mean()


class StaticAffine2dShifter(ModuleDict):
    def __init__(self, data_keys, input_channels=2, output_channels=2, bias=True, gamma_shifter=0):
        """
        Args:
            data_keys (list of str): keys of the shifter dictionary, correspond to the data_keys of the nnfabirk dataloaders
            gamma_shifter: weight of the regularizer

            See docstring of base class for the other arguments.
        """
        super().__init__()
        self.gamma_shifter = gamma_shifter
        for k in data_keys:
            self.add_module(k, StaticAffine2d(input_channels, output_channels, bias=bias))

    def initialize(self, bias=None):
        for k in self:
            if bias is not None:
                self[k].initialize(bias=bias[k])
            else:
                self[k].initialize()

    def regularizer(self, data_key):
        return self[data_key].regularizer() * self.gamma_shifter



import warnings

import numpy as np
import torch
import torch.nn as nn
from torch.nn import functional as F

# def laplace():
#     return np.array([[0.25, 0.5, 0.25], [0.5, -3.0, 0.5], [0.25, 0.5, 0.25]]).astype(np.float32)[None, None, ...]


def laplace():
    """
    Returns a 3x3 laplace filter.

    """
    return np.array([[0, -1, 0], [-1, 4, -1], [0, -1, 0]]).astype(np.float32)[None, None, ...]


def laplace5x5():
    """
    Returns a 5x5 LaplacianOfGaussians (LoG) filter.

    """
    return np.array([[0, 0, 1, 0, 0], [0, 1, 2, 1, 0], [1, 2, -16, 2, 1], [0, 1, 2, 1, 0], [0, 0, 1, 0, 0],]).astype(
        np.float32
    )[None, None, ...]


def laplace7x7():
    """
    Returns a 7x7 LaplacianOfGaussians (LoG) filter.

    """
    return np.array(
        [
            [0, 0, 1, 1, 1, 0, 0],
            [0, 1, 3, 3, 3, 1, 0],
            [1, 3, 0, -7, 0, 3, 1],
            [1, 3, -7, -24, -7, 3, 1],
            [1, 3, 0, -7, 0, 3, 1],
            [0, 1, 3, 3, 3, 1, 0],
            [0, 0, 1, 1, 1, 0, 0],
        ]
    ).astype(np.float32)[None, None, ...]


def laplace3d():
    l = np.zeros((3, 3, 3))
    l[1, 1, 1] = -6.0
    l[1, 1, 2] = 1.0
    l[1, 1, 0] = 1.0
    l[1, 0, 1] = 1.0
    l[1, 2, 1] = 1.0
    l[0, 1, 1] = 1.0
    l[2, 1, 1] = 1.0
    return l.astype(np.float32)[None, None, ...]


def gaussian2d(size, sigma=5, gamma=1, theta=0, center=(0, 0), normalize=True):
    """
    Returns a 2D Gaussian filter.

    Args:
        size (tuple of int, or int): Image height and width.
        sigma (float): std deviation of the Gaussian along x-axis. Default is 5..
        gamma (float): ratio between std devidation along x-axis and y-axis. Default is 1.
        theta (float): Orientation of the Gaussian (in ratian). Default is 0.
        center (tuple): The position of the filter. Default is center (0, 0).
        normalize (bool): Whether to normalize the entries. This is computed by
            subtracting the minimum value and then dividing by the max. Default is True.

    Returns:
        2D Numpy array: A 2D Gaussian filter.

    """

    sigma_x = sigma
    sigma_y = sigma / gamma

    xmax, ymax = (size, size) if isinstance(size, int) else size
    xmax, ymax = (xmax - 1) / 2, (ymax - 1) / 2
    xmin, ymin = -xmax, -ymax
    (y, x) = np.meshgrid(np.arange(ymin, ymax + 1), np.arange(xmin, xmax + 1))

    # shift the position
    y -= center[0]
    x -= center[1]

    # Rotation
    x_theta = x * np.cos(theta) + y * np.sin(theta)
    y_theta = -x * np.sin(theta) + y * np.cos(theta)

    gaussian = np.exp(-0.5 * (x_theta**2 / sigma_x**2 + y_theta**2 / sigma_y**2))

    if normalize:
        gaussian -= gaussian.min()
        gaussian /= gaussian.max()

    return gaussian.astype(np.float32)


class Laplace(nn.Module):
    """
    Laplace filter for a stack of data. Utilized as the input weight regularizer.
    """

    def __init__(self, padding=None, filter_size=3):
        """
        Laplace filter for a stack of data.

        Args:
            padding (int): Controls the amount of zero-padding for the convolution operation
                            Default is half of the kernel size (recommended)

        Attributes:
            filter (2D Numpy array): 3x3 Laplace filter.
            padding_size (int): Number of zeros added to each side of the input image
                before convolution operation.
        """
        super().__init__()
        if filter_size == 3:
            kernel = laplace()
        elif filter_size == 5:
            kernel = laplace5x5()
        elif filter_size == 7:
            kernel = laplace7x7()

        self.register_buffer("filter", torch.from_numpy(kernel))
        self.padding_size = self.filter.shape[-1] // 2 if padding is None else padding

    def forward(self, x):
        return F.conv2d(x, self.filter, bias=None, padding=self.padding_size)


class LaplaceL2(nn.Module):
    """
    Laplace regularizer for a 2D convolutional layer. Unnormalized, not recommended to use.
        Use LaplaceL2norm instead.

        Args:
            padding (int): Controls the amount of zero-padding for the convolution operation.

        Attributes:
            laplace (Laplace): Laplace convolution object. The output is the result of
                convolving an input image with laplace filter.

    """

    def __init__(self, padding=None, filter_size=3):
        super().__init__()
        self.laplace = Laplace(padding=padding, filter_size=filter_size)
        warnings.warn("LaplaceL2 Regularizer is deprecated. Use LaplaceL2norm instead.")

    def forward(self, x, avg=True):
        agg_fn = torch.mean if avg else torch.sum

        oc, ic, k1, k2 = x.size()
        return agg_fn(self.laplace(x.view(oc * ic, 1, k1, k2)).pow(2)) / 2


class LaplaceL2norm(nn.Module):
    """
    Normalized Laplace regularizer for a 2D convolutional layer.
        returns |laplace(filters)| / |filters|
    """

    def __init__(self, padding=None):
        super().__init__()
        self.laplace = Laplace(padding=padding)

    def forward(self, x, avg=False):
        agg_fn = torch.mean if avg else torch.sum

        oc, ic, k1, k2 = x.size()
        return agg_fn(self.laplace(x.view(oc * ic, 1, k1, k2)).pow(2)) / agg_fn(x.view(oc * ic, 1, k1, k2).pow(2))


class Laplace3d(nn.Module):
    """
    Laplace filter for a stack of data.
    """

    def __init__(self, padding=None):
        super().__init__()
        self.register_buffer("filter", torch.from_numpy(laplace3d()))

    def forward(self, x):
        return F.conv3d(x, self.filter, bias=None)


class LaplaceL23d(nn.Module):
    """
    Laplace regularizer for a 2D convolutional layer.
    """

    def __init__(self):
        super().__init__()
        self.laplace = Laplace3d()

    def forward(self, x):
        oc, ic, k1, k2, k3 = x.size()
        return self.laplace(x.view(oc * ic, 1, k1, k2, k3)).pow(2).mean() / 2


class FlatLaplaceL23d(nn.Module):
    """
    Laplace regularizer for a 2D convolutional layer.
    """

    def __init__(self):
        super().__init__()
        self.laplace = Laplace()

    def forward(self, x):
        oc, ic, k1, k2, k3 = x.size()
        assert k1 == 1, "time dimension must be one"
        return self.laplace(x.view(oc * ic, 1, k2, k3)).pow(2).mean() / 2


class LaplaceL1(nn.Module):
    """
    Laplace regularizer for a 2D convolutional layer.
    """

    def __init__(self, padding=0):
        super().__init__()
        self.laplace = Laplace(padding=padding)

    def forward(self, x, avg=True):
        agg_fn = torch.mean if avg else torch.sum

        oc, ic, k1, k2 = x.size()
        return agg_fn(self.laplace(x.view(oc * ic, 1, k1, k2)).abs())


class GaussianLaplaceL2Adaptive(nn.Module):
    """
    Laplace regularizer, with a Gaussian mask, for a 2D convolutional layer.
        Is flexible across kernel sizes, so that the regularizer can be applied to more
        than one layer, with changing kernel sizes
    """

    def __init__(self, padding=None, sigma=None):
        """
        Args:
            padding (int): Controls the amount of zero-padding for the convolution operation.
            sigma (float): std deviation of the Gaussian along x-axis. Default is calculated
                as the 1/4th of the minimum dimenison (height vs width) of the input.
        """
        super().__init__()
        self.laplace = Laplace(padding=padding)
        self.sigma = sigma

    def forward(self, x, avg=False):
        agg_fn = torch.mean if avg else torch.sum

        oc, ic, k1, k2 = x.size()
        sigma = self.sigma if self.sigma else min(k1, k2) / 4

        out = self.laplace(x.view(ic * oc, 1, k1, k2))
        out = out * (1 - torch.from_numpy(gaussian2d(size=(k1, k2), sigma=sigma)).expand(1, 1, k1, k2).to(x.device))

        return agg_fn(out.pow(2)) / agg_fn(x.view(oc * ic, 1, k1, k2).pow(2))


class GaussianLaplaceL2(nn.Module):
    """
    Laplace regularizer, with a Gaussian mask, for a single 2D convolutional layer.

    """

    def __init__(self, kernel, padding=None):
        """
        Args:
            kernel: Size of the convolutional kernel of the filter that is getting regularized
            padding (int): Controls the amount of zero-padding for the convolution operation.
        """
        super().__init__()

        self.laplace = Laplace(padding=padding)
        self.kernel = (kernel, kernel) if isinstance(kernel, int) else kernel
        sigma = min(*self.kernel) / 4
        self.gaussian2d = torch.from_numpy(gaussian2d(size=(*self.kernel,), sigma=sigma))

    def forward(self, x, avg=False):
        agg_fn = torch.mean if avg else torch.sum

        oc, ic, k1, k2 = x.size()
        out = self.laplace(x.view(oc * ic, 1, k1, k2))
        out = out * (1 - self.gaussian2d.expand(1, 1, k1, k2).to(x.device))

        return agg_fn(out.pow(2)) / agg_fn(x.view(oc * ic, 1, k1, k2).pow(2))



import warnings

import numpy as np
import torch
from torch import nn as nn
from torch.nn import Parameter
from torch.nn import functional as F

from .base import Readout


class Pyramid(nn.Module):
    _filter_dict = {
        "gauss5x5": np.float32(
            [
                [0.003765, 0.015019, 0.023792, 0.015019, 0.003765],
                [0.015019, 0.059912, 0.094907, 0.059912, 0.015019],
                [0.023792, 0.094907, 0.150342, 0.094907, 0.023792],
                [0.015019, 0.059912, 0.094907, 0.059912, 0.015019],
                [0.003765, 0.015019, 0.023792, 0.015019, 0.003765],
            ]
        ),
        "gauss3x3": np.float32([[1 / 16, 1 / 8, 1 / 16], [1 / 8, 1 / 4, 1 / 8], [1 / 16, 1 / 8, 1 / 16]]),
        "laplace5x5": np.outer(np.float32([1, 4, 6, 4, 1]), np.float32([1, 4, 6, 4, 1])) / 256,
    }

    def __init__(self, scale_n=4, type="gauss5x5", downsample=True):
        """
        Setup Laplace image pyramid
        Args:
            scale_n: number of Laplace pyramid layers to construct
            type: type of Gaussian filter used in pyramid construction. Valid options are: 'gauss5x5', 'gauss3x3', and 'laplace5x5'
            downsample: whether to downsample the image in each layer. Defaults to True
        """
        super().__init__()
        self.type = type
        self.downsample = downsample
        h = self._filter_dict[type]
        self.register_buffer("filter", torch.from_numpy(h))
        self.scale_n = scale_n
        self._kern = h.shape[0]
        self._pad = self._kern // 2
        self._filter_cache = None

    def lap_split(self, img):
        N, c, h, w = img.size()
        if self._filter_cache is not None and self._filter_cache.size(0) == c:
            filter = self._filter_cache
        else:
            filter = self.filter.expand(c, 1, self._kern, self._kern).contiguous()
            self._filter_cache = filter

        # the necessary output padding depends on even/odd of the dimension
        output_padding = (h + 1) % 2, (w + 1) % 2

        smooth = F.conv2d(img, filter, padding=self._pad, groups=c)
        if self.downsample:
            lo = smooth[:, :, ::2, ::2]
            lo2 = 4 * F.conv_transpose2d(
                lo,
                filter,
                stride=2,
                padding=self._pad,
                output_padding=output_padding,
                groups=c,
            )
        else:
            lo = lo2 = smooth

        hi = img - lo2

        return lo, hi

    def forward(self, img, **kwargs):
        levels = []
        for i in range(self.scale_n):
            img, hi = self.lap_split(img)
            levels.append(hi)
        levels.append(img)
        return levels

    def __repr__(self):
        return "Pyramid(scale_n={scale_n}, padding={_pad}, downsample={downsample}, type={type})".format(
            **self.__dict__
        )


class PointPyramid2d(Readout):
    def __init__(
        self,
        in_shape,
        outdims,
        scale_n,
        positive,
        bias,
        init_range,
        downsample,
        type,
        align_corners=True,
        mean_activity=None,
        feature_reg_weight=1.0,
        gamma_readout=None,  # depricated, use feature_reg_weight instead
        **kwargs,
    ):
        super().__init__()
        self.in_shape = in_shape
        c, w, h = in_shape
        self.outdims = outdims
        self.positive = positive
        self.feature_reg_weight = self.resolve_deprecated_gamma_readout(feature_reg_weight, gamma_readout)
        self.mean_activity = mean_activity
        self.gauss_pyramid = Pyramid(scale_n=scale_n, downsample=downsample, type=type)
        self.grid = Parameter(torch.Tensor(1, outdims, 1, 2))
        self.features = Parameter(torch.Tensor(1, c * (scale_n + 1), 1, outdims))

        if bias:
            bias = Parameter(torch.Tensor(outdims))
            self.register_parameter("bias", bias)
        else:
            self.register_parameter("bias", None)
        self.init_range = init_range
        self.align_corners = align_corners
        self.initialize(mean_activity)

    def initialize(self, mean_activity=None):
        if mean_activity is None:
            mean_activity = self.mean_activity
        self.grid.data.uniform_(-self.init_range, self.init_range)
        self.features.data.fill_(1 / self.in_shape[0])
        if self.bias is not None:
            self.initialize_bias(mean_activity=mean_activity)

    def group_sparsity(self, group_size):
        f = self.features.size(1)
        n = f // group_size
        ret = 0
        for chunk in range(0, f, group_size):
            ret = ret + (self.features[:, chunk : chunk + group_size, ...].pow(2).mean(1) + 1e-12).sqrt().mean() / n
        return ret

    def feature_l1(self, reduction="sum", average=None):
        return self.apply_reduction(self.features.abs(), reduction=reduction, average=average)

    def regularizer(self, reduction="sum", average=None):
        return self.feature_l1(reduction=reduction, average=average) * self.feature_reg_weight

    def forward(self, x, shift=None, **kwargs):
        if self.positive:
            self.features.data.clamp_min_(0)
        self.grid.data = torch.clamp(self.grid.data, -1, 1)
        N, c, w, h = x.size()
        m = self.gauss_pyramid.scale_n + 1
        feat = self.features.view(1, m * c, self.outdims)

        if shift is None:
            grid = self.grid.expand(N, self.outdims, 1, 2)
        else:
            grid = self.grid.expand(N, self.outdims, 1, 2) + shift[:, None, None, :]

        pools = [F.grid_sample(xx, grid, align_corners=self.align_corners) for xx in self.gauss_pyramid(x)]
        y = torch.cat(pools, dim=1).squeeze(-1)
        y = (y * feat).sum(1).view(N, self.outdims)

        if self.bias is not None:
            y = y + self.bias
        return y

    def __repr__(self):
        c, w, h = self.in_shape
        r = self.__class__.__name__ + " (" + "{} x {} x {}".format(c, w, h) + " -> " + str(self.outdims) + ")"
        if self.bias is not None:
            r += " with bias"

        for ch in self.children():
            r += "  -> " + ch.__repr__() + "\n"
        return r



from torch import nn


class DepthSeparableConv2d(nn.Sequential):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        bias=True,
    ):
        super().__init__()
        self.add_module("in_depth_conv", nn.Conv2d(in_channels, out_channels, 1, bias=bias))
        self.add_module(
            "spatial_conv",
            nn.Conv2d(
                out_channels,
                out_channels,
                kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                bias=bias,
                groups=out_channels,
            ),
        )
        self.add_module("out_depth_conv", nn.Conv2d(out_channels, out_channels, 1, bias=bias))
